// vincia03-root.cc is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to study light-flavor fragmentation 
// at LEP (compared to L3 udsc measurements), using VINCIA's ROOT-based 
// runtime display. 
// Uses VINCIA for showering and matching and PYTHIA 8 for all else
// Copyright (c) 2017 Peter Skands

// PYTHIA and VINCIA include files and namespaces
#include "Vincia/Vincia.h"
using namespace Pythia8;
using namespace Vincia;
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot/VinciaRoot.h"
#include "VinciaRoot/PythiaRoot.h"

// Global on/off switches for optional additional plots:
// 1) pTin and pTout distributions
const bool OPTshowPTinout = true;
// 2) identified-particle spectra
const bool OPTshowParticleSpectra = false;
// 3) identified-particle parent composition
const bool OPTshowParticleComposition = false;
// 4) initialise from saved runs?
const bool OPTinitialiseFromSaved = false;

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects. 
  Pythia pythia; 
  VinciaPlugin vincia(&pythia);

  // Read user settings from file
  pythia.readFile("vincia03.cmnd");
 
  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Shorthand for event record
  Event* eventPtr = &pythia.event;

  // Some useful pointers
  Settings* settingsPtr   = vincia.settingsPtr;        // to settings database  

  // Extract settings to be used in the main program.  
  double eCM              = settingsPtr->parm("Beams:eCM");
  int    nEvent           = settingsPtr->mode("Main:numberOfEvents");  

  // Check if hadronisation is on (used to make plot labels)
  bool   hadronLevel      = settingsPtr->flag("HadronLevel:all");

  // Vincia settings
  bool   vinciaOn         = settingsPtr->flag("Vincia");

  // Data path directory, extract from VINCIADATA
  string VINCIA_SHARE     = vincia.VINCIA_SHARE;
  string VINCIA_VAL       = VINCIA_SHARE+"validation/";
  string VINCIA_DAT       = VINCIA_SHARE+"data/";
  
  //************************************************************************
  // VinciaRoot
  // Create VinciaRoot object and set plot labels, legends, etc
  // Also define names of files to save and read histograms to/from
  //************************************************************************

  // Create VinciaRoot object and give it a pointer to read its settings
  VinciaRoot vroot(settingsPtr);
  vroot.init();

  // Set default plot properties
  vroot.setLogY();                         // Logarithmic Y axes
  vroot.setRatioRange("all",0.5,1.5);      // Range on ratio plots (default is [0.5,1.5])
  vroot.setShowLegendChi2("all",true);    // Compute and display chi2(5%) values?
  vroot.setLeftLabel("all","ee#rightarrowhadrons"); // Left-hand top label
  vroot.setRightLabel("all","91.2 GeV");            // Right-hand top label

  // Set label for this run
  string mcLabel = (vinciaOn) ? "Vincia" : "Pythia";
  if (!hadronLevel) mcLabel += " partons";
  // Set VinciaRoot plot labels, legends, etc
  vroot.setLegend("all",mcLabel);
  
  // Compare current run to files with the following prefix(es)
  vector<string> mcFileRead;
  mcFileRead.push_back(string(VINCIA_VAL+"vincia2/vincia03").c_str());
  //mcFileRead.push_back(string(VINCIA_VAL+"vincia1/vincia03").c_str());
  mcFileRead.push_back(string(VINCIA_VAL+"pythia/vincia03").c_str());
  //mcFileRead.push_back("last/vincia03");

  // Prefix of rescue files (to initialise from last run)
  string rescuePrefix = "vincia03-";
  
  // Next event at which VinciaRoot will save displays/histograms
  int iSave = vroot.nextSave();

  //************************************************************************
  // Initialize VINCIA (automatically initializes PYTHIA) 
  //************************************************************************

  // Initialize
  vincia.init();

  //************************************************************************
  // Book Histograms
  //************************************************************************

  // Vincia uncertainty bands: number of variations expected 
  int nWeights = max(1,vincia.nWeights());

  // Default plot sizes
  double plotSizeX = 360;
  double plotSizeY = 420;

  // Histogram bin and range variables
  int nBins;
  double xLo, xHi;

  // Book alphaS histogram
  nBins = 30;
  xLo   = 0.;
  xHi   = 3.0;
  TH1D alphaS("alphaS","#alpha_{s}",nBins,xLo,xHi);
  TH1D alphaStwo(alphaS);
  TH1D alphaShalf(alphaS);
  TH1D alphaS1(alphaS);
  TH1D alphaSpy("alphaSpy","#alpha_{s}",nBins,xLo,xHi);
  TH1D alphaScmw("alphaScmw","#alpha_{s}",nBins,xLo,xHi);
  vector<TH1D*> vAlphaS;
  vAlphaS.push_back(&alphaS);
  vAlphaS.push_back(&alphaShalf);
  vAlphaS.push_back(&alphaStwo);
  // Set properties 
  alphaS.SetMinimum(0.05);
  alphaS.SetMaximum(2.0);
  alphaS.GetXaxis()->SetTitle("Log10(p_{T}) [GeV]");
  vroot.createDisplay("alphaS",vAlphaS);  
  vroot.setShowRelUnc("alphaS",false);
  vroot.setLogY("alphaS",true);
  vroot.setLeftLabel("alphaS","");
  vroot.setRightLabel("alphaS","");
  vroot.setMCnames("alphaS","");
  vroot.setNormalize("alphaS",false);
  vroot.setRatioRange("alphaS",0.5,1.5);
  
  // Fill histogram
  if (vinciaOn) {
    bool useCMW            = settingsPtr->flag("Vincia:useCMW");
    AlphaStrong* alphaSptr = &vincia.vinCom.alphaStrong;
    if (useCMW) alphaSptr  = &vincia.vinCom.alphaStrongCMW;
    AlphaStrong alphaPyDef;
    alphaPyDef.init(0.1365, 1, 5, false);
    AlphaStrong alpha2Lcmw;
    alpha2Lcmw.init(0.1188, 2, 5, true);
    vroot.addTH1DPtr("alphaS",&alphaSpy);
    vroot.addTH1DPtr("alphaS",&alphaScmw);
    for ( int iBin=1; iBin<=nBins; ++iBin ) {
      double logQ = xLo + (iBin-0.5)*(xHi-xLo)/nBins;
      double pT   = pow(10.,logQ);
      double kPT2      = pow2(settingsPtr->parm("Vincia:alphaSkMuF") * pT);
      double mu2freeze = pow2(settingsPtr->parm("Vincia:alphaSmuFreeze"));
      double as        = alphaSptr->alphaS(kPT2 + mu2freeze);
      double asTwo     = alphaSptr->alphaS(4*(kPT2+mu2freeze));
      double asHalf    = alphaSptr->alphaS((kPT2+mu2freeze)/4.);
      alphaS.SetBinContent(iBin,as);
      alphaStwo.SetBinContent(iBin,asTwo);
      alphaShalf.SetBinContent(iBin,asHalf);
      alphaS.SetBinError(iBin,0.);	
      alphaStwo.SetBinError(iBin,0.);	
      alphaShalf.SetBinError(iBin,0.);	
      double asPy   = alphaPyDef.alphaS(pow2(pT));
      alphaSpy.SetBinContent(iBin,asPy);
      alphaSpy.SetBinError(iBin,0.);
      double as2Lcmw   = alpha2Lcmw.alphaS(pow2(pT));
      alphaScmw.SetBinContent(iBin,as2Lcmw);
      alphaScmw.SetBinError(iBin,0.);
    }    
    vector<string> legendForAlphaS;
    legendForAlphaS.push_back("Vincia");
    legendForAlphaS.push_back("Pythia");
    legendForAlphaS.push_back("CMW 0.1188 2L");
    vroot.setLegend("alphaS",legendForAlphaS);
    vroot.updateDisplay("alphaS");    
  }


  // weights (for uncertainty bands)
  vector<TH1D*> Vweights;
  for (int i=0; i<nWeights; ++i)
    Vweights.push_back(new TH1D(string("weights"+num2str(i,0)).c_str(),
      "Event Weights",50, -2.0, 4.0));
  Vweights[0]->SetMinimum(0.005);
  Vweights[0]->SetMaximum(50.0);
  Vweights[0]->GetXaxis()->SetTitle("log_{10}(w)");
  Vweights[0]->GetYaxis()->SetTitle("Rate");
  if (nWeights >= 1) {
    vroot.createDisplay("weights",Vweights[0]);
    for (int i=1; i<nWeights; ++i)
      vroot.addTH1DPtr("weights",Vweights[i]);
    vroot.setNormalize("weights");
    vroot.setShowRatioPad("weights",false);
    vector<string> bandLabels;
    bandLabels.push_back("0");
    bandLabels.push_back("1");
    bandLabels.push_back("2");
    bandLabels.push_back("3");
    bandLabels.push_back("4");
    vroot.setLegend("weights",bandLabels);    
  }
  
  // nPartons
  TH1D *HnPartons=new TH1D("nPartons","N_{Partons}(Q_{had})", 50, -0.5, 49.5);
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(HnPartons,rescuePrefix+"nPartons.dat");
  HnPartons->SetMinimum(0.0005);
  HnPartons->SetMaximum(0.5);
  vroot.createDisplay("nPartons",HnPartons);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
  vroot.setLogY(string("nPartons"),true);
  vroot.setNormalize("nPartons");

  // Quark Pairs
  nBins = 10;
  xLo   = -0.5;
  xHi   = 9.5;
  vector<TH1D*> VnQuarks;
  VnQuarks.push_back(new TH1D("nQuarks","N_{Quarks}(Q_{had})",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VnQuarks[0],rescuePrefix+"nQuarks.dat");
  VnQuarks[0]->GetXaxis()->SetTitle("N_{Q}");
  VnQuarks[0]->GetYaxis()->SetTitle("P(N_{Q})");
  VnQuarks[0]->SetMinimum(0.00005);
  VnQuarks[0]->SetMaximum(5.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VnQuarks.push_back(new TH1D(string("nQuarks"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.createDisplay("nQuarks",VnQuarks,320,360);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    vroot.addHisFile("nQuarks",mcFileRead[iRead]+"-nQuarks.dat");
  vroot.setLogY("nQuarks");
  vroot.setNormalize("nQuarks");

  // Number of final-state emissions
  nBins = 51;
  xLo   = -0.5;
  xHi   = 50.5;
  vector<TH1D*> VnFSR;
  VnFSR.push_back(new TH1D("nFSR","N_{FSR}(Q_{had})", nBins, xLo, xHi));
  VnFSR[0]->GetXaxis()->SetTitle("N_{FSR}");
  VnFSR[0]->GetYaxis()->SetTitle("P(N_{FSR})");
  VnFSR[0]->SetMinimum(0.0005);
  VnFSR[0]->SetMaximum(0.5);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VnFSR.push_back(new TH1D(string("nFSR"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.createDisplay("nFSR",VnFSR);
  vroot.setLogY(string("nFSR"),true);
  vroot.setNormalize("nFSR");

  // Mesons
  nBins = 16;
  xLo   = -0.5;
  xHi   = xLo + nBins;
  vector<TH1D*> VMesons;
  VMesons.push_back(new TH1D("Mesons","Meson Fractions",nBins,xLo,xHi));
  VMesons[0]->GetYaxis()->SetTitle("#frac{<n>}{<n_{ch}>}");
  VMesons[0]->GetXaxis()->SetBinLabel(1,"#pi^{#pm}");
  VMesons[0]->GetXaxis()->SetBinLabel(2,"#pi^{0}");
  VMesons[0]->GetXaxis()->SetBinLabel(3,"K^{#pm}");
  VMesons[0]->GetXaxis()->SetBinLabel(4,"#eta");
  VMesons[0]->GetXaxis()->SetBinLabel(5,"#eta'");
  VMesons[0]->GetXaxis()->SetBinLabel(6,"#rho^{#pm}");
  VMesons[0]->GetXaxis()->SetBinLabel(7,"#rho^{0}"); 
  VMesons[0]->GetXaxis()->SetBinLabel(8,"#omega");
  VMesons[0]->GetXaxis()->SetBinLabel(9,"K*^{#pm}");
  VMesons[0]->GetXaxis()->SetBinLabel(10,"#phi");
  VMesons[0]->GetXaxis()->SetBinLabel(11,"R_{K*/K}");
  VMesons[0]->GetXaxis()->SetBinLabel(12,"R_{#phi/K*}");
  VMesons[0]->GetXaxis()->SetBinLabel(13,"R_{#phi/K}");
  VMesons[0]->GetXaxis()->SetBinLabel(14,"R_{#phi/#rho}");
  VMesons[0]->GetXaxis()->SetBinLabel(15,"R_{#phi/#omega}");
  VMesons[0]->GetXaxis()->SetBinLabel(16,"R_{#phi/#pi^{#pm}}");
  VMesons[0]->GetXaxis()->SetLabelOffset(0.04);
  VMesons[0]->GetXaxis()->SetLabelSize(0.07);
  VMesons[0]->SetMaximum(20.01); 
  VMesons[0]->SetMinimum(0.001); 
  if (hadronLevel) {
    vroot.createDisplay("Mesons",VMesons,1.15*plotSizeX,plotSizeY);
    vroot.setNormalize("Mesons",false);
    //vroot.setRatioRange("Mesons",0.5,1.5);
    vroot.setDataFile("Mesons",string(VINCIA_DAT+"LEP91-mesons.dat").c_str(),20,40,80);  
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("Mesons",mcFileRead[iRead]+"-Mesons.dat");
  }

  // Baryons
  nBins = 11;
  xLo   = -0.5;
  xHi   = 10.5;
  vector<TH1D*> VBaryons;
  VBaryons.push_back(new TH1D("Baryons","Baryon Fractions",nBins, xLo, xHi));  
  VBaryons[0]->GetYaxis()->SetTitle("#frac{<n>}{<n_{ch}>}");
  VBaryons[0]->GetXaxis()->SetBinLabel(1,"p^{#pm}");
  VBaryons[0]->GetXaxis()->SetBinLabel(2,"#Lambda");
  VBaryons[0]->GetXaxis()->SetBinLabel(3,"#Sigma^{#pm}");
  VBaryons[0]->GetXaxis()->SetBinLabel(4,"#Sigma^{0}");
  VBaryons[0]->GetXaxis()->SetBinLabel(5,"#Xi^{#pm}");
  VBaryons[0]->GetXaxis()->SetBinLabel(6,"#Delta^{++}");
  VBaryons[0]->GetXaxis()->SetBinLabel(7,"#Sigma^{#pm}*");
  VBaryons[0]->GetXaxis()->SetBinLabel(8,"#Xi^{0}*");
  VBaryons[0]->GetXaxis()->SetBinLabel(9,"#Omega");
  VBaryons[0]->GetXaxis()->SetBinLabel(10,"R_{#Lambda/p}");
  VBaryons[0]->GetXaxis()->SetBinLabel(11,"R_{#Lambda/K}");
  VBaryons[0]->GetXaxis()->SetLabelOffset(0.04);
  VBaryons[0]->GetXaxis()->SetLabelSize(0.07);
  VBaryons[0]->SetMaximum(10.0); 
  VBaryons[0]->SetMinimum(0.00001); 
  if (hadronLevel) {
    vroot.createDisplay("Baryons",VBaryons,plotSizeX,plotSizeY);
    vroot.setNormalize("Baryons",false);
    //vroot.setRatioRange("Baryons",0.5,1.5);
    vroot.setDataFile("Baryons",string(VINCIA_DAT+"LEP91-baryons.dat").c_str(),20,40,80);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("Baryons",mcFileRead[iRead]+"-Baryons.dat");
  }    

  // Heavy Quarks
  nBins =  9;
  xLo   = -0.5;
  xHi   = 8.5;
  vector<TH1D*> VCharm;
  VCharm.push_back(new TH1D("Charm","Charm Rates",nBins, xLo, xHi));
  VCharm[0]->GetYaxis()->SetTitle("<n> or BR(Z#rightarrowX)");
  VCharm[0]->GetXaxis()->SetBinLabel(1,"D^{#pm}");
  VCharm[0]->GetXaxis()->SetBinLabel(2,"D^{0}");
  VCharm[0]->GetXaxis()->SetBinLabel(3,"D^{*#pm}");
  VCharm[0]->GetXaxis()->SetBinLabel(4,"D_{s}^{#pm}");
  VCharm[0]->GetXaxis()->SetBinLabel(5,"#Lambda_{c}^{+}X");
  VCharm[0]->GetXaxis()->SetBinLabel(6,"g#rightarrowc#bar{c}");
  VCharm[0]->GetXaxis()->SetBinLabel(7,"J/#psi");
  VCharm[0]->GetXaxis()->SetBinLabel(8,"#chi_{c1}");
  VCharm[0]->GetXaxis()->SetBinLabel(9,"#psi_{3685}");
  VCharm[0]->GetXaxis()->SetLabelOffset(0.04);
  VCharm[0]->GetXaxis()->SetLabelSize(0.064);
  VCharm[0]->SetMaximum(50.0); 
  VCharm[0]->SetMinimum(0.0001); 
  if (hadronLevel) {
    vroot.createDisplay("Charm",VCharm,plotSizeX,plotSizeY);
    vroot.setNormalize("Charm",false);
    vroot.setRatioRange("Charm",0.0,1.99);
    vroot.setDataFile("Charm",string(VINCIA_DAT+"LEP-91-charm.dat").c_str(),20,40,60,70);  
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("Charm",mcFileRead[iRead]+"-Charm.dat");
  }

  nBins =  8;
  xLo   = -0.5;
  xHi   = 7.5;
  vector<TH1D*> VBeauty;
  VBeauty.push_back(new TH1D("Beauty","Beauty Rates",nBins, xLo, xHi));
  VBeauty[0]->GetYaxis()->SetTitle("<n> or BR(Z#rightarrowX)");
  VBeauty[0]->GetXaxis()->SetBinLabel(1,"B^{+}X");
  VBeauty[0]->GetXaxis()->SetBinLabel(2,"B^{#pm0}");
  VBeauty[0]->GetXaxis()->SetBinLabel(3,"B^{*}_{uds}");
  VBeauty[0]->GetXaxis()->SetBinLabel(4,"B_{s}^{0}X");
  VBeauty[0]->GetXaxis()->SetBinLabel(5,"B_{bqq}X");
  VBeauty[0]->GetXaxis()->SetBinLabel(6,"g#rightarrowb#bar{b}");
  VBeauty[0]->GetXaxis()->SetBinLabel(7,"4b");
  VBeauty[0]->GetXaxis()->SetBinLabel(8,"#Upsilon(#times10)");
  VBeauty[0]->GetXaxis()->SetLabelOffset(0.04);
  VBeauty[0]->GetXaxis()->SetLabelSize(0.064);
  VBeauty[0]->SetMaximum(50.0); 
  VBeauty[0]->SetMinimum(0.00001); 
  if (hadronLevel) {
    vroot.createDisplay("Beauty",VBeauty,plotSizeX,plotSizeY);
    vroot.setNormalize("Beauty",false);
    vroot.setRatioRange("Beauty",0.0,1.99);
    vroot.setDataFile("Beauty",string(VINCIA_DAT+"LEP-91-beauty.dat").c_str(),20,40,60,70);  
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("Beauty",mcFileRead[iRead]+"-Beauty.dat");
  }

  // nCharged (udsc)
  nBins = 61;
  xLo   = -0.5;
  xHi   = 60.5;
  vector<TH1D*> VnCharged;
  VnCharged.push_back(new TH1D("nCh","Charged Multiplicity (udsc)",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VnCharged[0],rescuePrefix+"Nch.dat");
  VnCharged[0]->GetXaxis()->SetTitle("n_{ch}");
  VnCharged[0]->GetYaxis()->SetTitle("Probability(n_{ch})");
  VnCharged[0]->SetMinimum(0.000001);
  VnCharged[0]->SetMaximum(200.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VnCharged.push_back(new TH1D(string("nCh"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel) {
    vroot.createDisplay("Nch",VnCharged,plotSizeX,plotSizeY);
    vroot.setZeroSkipping("Nch",true);
    vroot.setNormalize("Nch",true);
    vroot.setDataFile("Nch",string(VINCIA_DAT+"L3-91-Nch-udsc.dat").c_str(),20,20,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("Nch",mcFileRead[iRead]+"-Nch.dat");
    vroot.setLogY("Nch");
  }

  // Ln(1/x) (udsc)
  nBins = 45;
  xLo   = -0.2;
  xHi   = 8.8;
  vector<TH1D*> Vlnx;
  Vlnx.push_back(new TH1D("LnXInv","Charged Momentum Fraction (udsc)",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(Vlnx[0],rescuePrefix+"Lnx.dat");
  double wLnx = nBins/(xHi-xLo);
  Vlnx[0]->GetXaxis()->SetTitle("|Ln(x_{p})|");
  Vlnx[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/d|Ln(x_{p})|");
  Vlnx[0]->SetMinimum(0.0004);
  Vlnx[0]->SetMaximum(50.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    Vlnx.push_back(new TH1D(string("LnXInv"+num2str(iVar,0)).c_str(),
			  " ",nBins,xLo,xHi));
  if (hadronLevel) {
    vroot.createDisplay("Lnx",Vlnx,plotSizeX,plotSizeY);
    vroot.setNormalize("Lnx",true);
    vroot.setDataFile("Lnx",string(VINCIA_DAT+"L3-91-LogInvx-udsc.dat").c_str(),10,30,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("Lnx",mcFileRead[iRead]+"-Lnx.dat");
    vroot.setLogY("Lnx");
    vroot.setRebinLikeData("Lnx");
  }

  // Particle composition vs Ln(1/x) (udsc)
  vector<TH1D*> VidLnxSum;
  vector<TH1D*> VidLnx;
  VidLnxSum.push_back(new TH1D("IdLnXInvSum1", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VidLnxSum.push_back(new TH1D("IdLnXInvSum2", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VidLnxSum.push_back(new TH1D("IdLnXInvSum3", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VidLnxSum.push_back(new TH1D("IdLnXInvSum4", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VidLnx.push_back(new TH1D("IdLnXInv1",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VidLnx.push_back(new TH1D("IdLnXInv2",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VidLnx.push_back(new TH1D("IdLnXInv3",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VidLnx.push_back(new TH1D("IdLnXInv4",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VidLnx[0]->GetXaxis()->SetTitle("|Ln(x_{p})|");
  VidLnx[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ID}/d|Ln(x_{p})|");
  VidLnx[0]->SetMinimum(0.0);
  VidLnx[0]->SetMaximum(1.25);
  if (hadronLevel && OPTshowParticleComposition) { 
    vroot.createDisplay("LnxID",VidLnx[0],plotSizeX,plotSizeY);
    for (int i=1; i<int(VidLnx.size()); ++i) 
      vroot.addTH1DPtr("LnxID",VidLnx[i]);
    vroot.setLogY("LnxID",false);
    vroot.setNormalize("LnxID",false);
    vector<string> legendID;
    legendID.push_back("#pi^{#pm}");
    legendID.push_back("K^{#pm}");
    legendID.push_back("p^{#pm}");
    legendID.push_back("Other");
    vroot.setLegend("LnxID",legendID);
  }

  // Particle composition vs Ln(1/x) (udsc)
  vector<TH1D*> VkChLnxSum;
  vector<TH1D*> VkChLnx;
  nBins = 35;
  xLo   = -0.2;
  xHi   = 6.8;
  VkChLnxSum.push_back(new TH1D("KchLnXInvSum0", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnxSum.push_back(new TH1D("KchLnXInvSum1", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnxSum.push_back(new TH1D("KchLnXInvSum2", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnxSum.push_back(new TH1D("KchLnXInvSum3", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnxSum.push_back(new TH1D("KchLnXInvSum4", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnxSum.push_back(new TH1D("KchLnXInvSum5", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnxSum.push_back(new TH1D("KchLnXInvSum6", 
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnx.push_back(new TH1D("KchLnXInv0",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnx.push_back(new TH1D("KchLnXInv1",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnx.push_back(new TH1D("KchLnXInv2",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnx.push_back(new TH1D("KchLnXInv3",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnx.push_back(new TH1D("KchLnXInv4",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnx.push_back(new TH1D("KchLnXInv5",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnx.push_back(new TH1D("KchLnXInv6",
                   "Particle Composition vs Lnx (udsc)",nBins,xLo,xHi));
  VkChLnx[0]->GetXaxis()->SetTitle("|Ln(x_{p})|");
  VkChLnx[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{K}/d|Ln(x_{p})|");
  if (hadronLevel && OPTshowParticleComposition) {
    vroot.createDisplay("LnxKch",VkChLnx[0],plotSizeX,plotSizeY);
    for (int i=1; i<int(VkChLnx.size()); ++i) 
      vroot.addTH1DPtr("LnxKch",VkChLnx[i]);
    vroot.setLogY("LnxKch",false);
    vroot.setNormalize("LnxKch",false);
    vector<string> legendID;
    legendID.push_back("Total");
    legendID.push_back("s quarks");
    legendID.push_back("D and B mesons");
    legendID.push_back("K*");
    legendID.push_back("u and d quarks");
    legendID.push_back("c and b quarks");
    legendID.push_back("Other");
    vroot.setLegend("LnxKch",legendID);
    vroot.setRatioRange("LnxKch",0.0,1.0);
  }

  // x (all)
  nBins = 200;
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1D*> Vx;
  Vx.push_back(new TH1D("X","Charged Momentum Fraction",nBins,xLo,xHi));
  double wx = nBins/(xHi-xLo);
  Vx[0]->GetXaxis()->SetTitle("x_{p}");
  Vx[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dx_{p}");
  Vx[0]->SetMinimum(0.00005);
  Vx[0]->SetMaximum(40.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    Vx.push_back(new TH1D(string("X"+num2str(iVar,0)).c_str(),
			  " ",nBins,xLo,xHi));
  if (hadronLevel) {
    vroot.createDisplay("x",Vx,plotSizeX,plotSizeY);
    vroot.setNormalize("x",true);
    vroot.setDataFile("x",string(VINCIA_DAT+"Aleph-91-x.dat").c_str(),10,30,40,50,60);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("x",mcFileRead[iRead]+"-x.dat");
    vroot.setLogY("x");
    vroot.setRebinLikeData("x");
  }

  // Auxiliary histogram for storing sum(|p|)
  vector<TH1D*> VpSum;
  VpSum.push_back(new TH1D("pSum","|p| vs Charged Momentum Fraction",nBins,xLo,xHi));
  
  // Rapidity wrt Thrust
  nBins = 24;
  xLo   = 0.0;
  xHi   = 6.0;
  vector<TH1D*> VyThr;
  VyThr.push_back(new TH1D("yThr","Charged Rapidity wrt Thrust",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VyThr[0],rescuePrefix+"yThr.dat");
  double wyThr = nBins/(xHi-xLo);
  VyThr[0]->GetXaxis()->SetTitle("|y|");
  VyThr[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/d|y|");
  VyThr[0]->SetMinimum(0.00005);
  VyThr[0]->SetMaximum(60.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VyThr.push_back(new TH1D(string("yThr"+num2str(iVar,0)).c_str(),
			  " ",nBins,xLo,xHi));
  vroot.createDisplay("yThr",VyThr,plotSizeX,plotSizeY);
  vroot.setNormalize("yThr",true);
  vroot.setDataFile("yThr",string(VINCIA_DAT+"Aleph-91-yThr.dat").c_str(),10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    vroot.addHisFile("yThr",mcFileRead[iRead]+"-yThr.dat");

  // x(K+-) from ALEPH, used to zoom in on the soft part of the spectrum
  nBins = 210;  // (needed due to non-uniform bins in data)
  xLo   = -5.25;
  xHi   = 0.;
  vector<TH1D*> VxKch;
  VxKch.push_back(new TH1D("xKch","K^{#pm} momentum fraction (Combined)",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VxKch[0],rescuePrefix+"xKch.dat");
  double wxKch = nBins/(xHi-xLo);
  VxKch[0]->GetXaxis()->SetTitle("ln(x_{p})");
  VxKch[0]->GetYaxis()->SetTitle("1/<n_{K}> dn_{K}/dln(x)");
  VxKch[0]->SetMinimum(0.0005);
  VxKch[0]->SetMaximum(60.);  
  for (int iVar=1; iVar<nWeights; iVar++) 
    VxKch.push_back(new TH1D(string("xKch"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    // Make this plot extra wide to show tiny bins near zero
    vroot.createDisplay("xKch",VxKch,plotSizeX,plotSizeY);
    vroot.setNormalize("xKch",true);
    vroot.setDataFile("xKch",string(VINCIA_DAT+"Aleph-Opal-Delphi-91-xK.dat").c_str(),13,33,43,50,60,71,72);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("xKch",mcFileRead[iRead]+"-xKch.dat");
    vroot.setRebinLikeData("xKch");
  }

  // x(phi) (heaviest meson spectrum)
  nBins = 200;  // (needed due to non-uniform bins in data)
  xLo   = 0.;
  xHi   = 1.;
  vector<TH1D*> VxPhi;
  VxPhi.push_back(new TH1D("xPhi","#phi momentum fraction",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VxPhi[0],rescuePrefix+"xPhi.dat");
  double wxPhi = nBins/(xHi-xLo);
  VxPhi[0]->GetXaxis()->SetTitle("x_{p}");
  VxPhi[0]->GetYaxis()->SetTitle("1/<n_{#phi}> dn_{#phi}/dx");
  VxPhi[0]->SetMinimum(0.01);
  VxPhi[0]->SetMaximum(20.);  
  for (int iVar=1; iVar<nWeights; iVar++) 
    VxPhi.push_back(new TH1D(string("xPhi"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    // Make this plot extra wide to show tiny bins near zero
    vroot.createDisplay("xPhi",VxPhi,plotSizeX,plotSizeY);
    vroot.setNormalize("xPhi",true);
    vroot.setDataFile("xPhi",string(VINCIA_DAT+"Aleph-91-xphi.dat").c_str(),
                      10,20,30,40,80);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("xPhi",mcFileRead[iRead]+"-xPhi.dat");
    vroot.setRatioRange("xPhi",0.5,1.99);
    vroot.setRebinLikeData("xPhi");
  }
  
  // x(protons) ALEPH + SLD combined
  nBins = 253;  // (needed due to non-uniform bins in data)
  xLo   = -5;
  xHi   = 0.0;
  vector<TH1D*> Vxp;
  Vxp.push_back(new TH1D("xp","p+#bar{p} momentum fraction (Combined)",nBins,xLo,xHi));
  double wxp = nBins/(xHi-xLo);
  Vxp[0]->GetXaxis()->SetTitle("x_{p}");
  Vxp[0]->GetYaxis()->SetTitle("1/<n_{p}> dn_{p}/dx");
  Vxp[0]->SetMinimum(0.0001);
  Vxp[0]->SetMaximum(500.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    Vxp.push_back(new TH1D(string("xp"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    vroot.createDisplay("xProton",Vxp,plotSizeX,plotSizeY);
    vroot.setNormalize("xProton",true);
    vroot.setDataFile("xProton",string(VINCIA_DAT+"ALEPH-SLD-91-xp.dat").c_str(),13,33,43,50,60,
                      71,72);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("xProton",mcFileRead[iRead]+"-xProton.dat");
    vroot.setRebinLikeData("xProton");
  }

  // Lnx(Lambda)
  nBins = 25;  // (needed due to non-uniform bins in data)
  xLo   = 0;
  xHi   = 5.0;
  vector<TH1D*> VxLam;
  VxLam.push_back(new TH1D("xLam","#Lambda^{0} momentum fraction (-ln[x])",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VxLam[0],rescuePrefix+"xLam.dat");
  double wxLam = nBins/(xHi-xLo);
  VxLam[0]->GetXaxis()->SetTitle("#xi_{p}");
  VxLam[0]->GetYaxis()->SetTitle("1/<n_{#Lambda}> dn_{#Lambda}/d#xi");
  VxLam[0]->SetMinimum(0.0);
  VxLam[0]->SetMaximum(0.7);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VxLam.push_back(new TH1D(string("xLam"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    vroot.createDisplay("xLambda",VxLam,plotSizeX,plotSizeY);
    vroot.setNormalize("xLambda",true);
    vroot.setDataFile("xLambda",string(VINCIA_DAT+"Aleph-91-LogInvxL.dat").c_str(),10,30,40,50,60);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("xLambda",mcFileRead[iRead]+"-xLambda.dat");
    vroot.setLogY("xLambda",false);
  }

  // x(Xi)
  nBins = 55;  // (needed due to non-uniform bins in data)
  xLo   = 0;
  xHi   = 0.55;
  vector<TH1D*> VxXi;
  VxXi.push_back(new TH1D("xXi","#Xi^{#pm} momentum fraction",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VxXi[0],rescuePrefix+"xXi.dat");
  double wxXi = nBins/(xHi-xLo);
  VxXi[0]->GetXaxis()->SetTitle("x_{E}");
  VxXi[0]->GetYaxis()->SetTitle("1/<n_{#Xi}> dn_{#Xi}/dx");
  VxXi[0]->SetMinimum(0.07);
  VxXi[0]->SetMaximum(50.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VxXi.push_back(new TH1D(string("xXi"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    vroot.createDisplay("xXi",VxXi,plotSizeX,plotSizeY);
    vroot.setNormalize("xXi",true);
    vroot.setDataFile("xXi",string(VINCIA_DAT+"Aleph-91-xXi.dat").c_str(),10,30,40,80,0);
    vroot.setLogY("xXi");
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("xXi",mcFileRead[iRead]+"-xXi.dat");
    vroot.setRebinLikeData("xXi");
  }

  // x(D*)
  nBins = 20;  
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1D*> VxDstarCh;
  VxDstarCh.push_back(new TH1D("xDstarCh","D^{*#pm} energy fraction (x_{E}>0.1)",nBins,xLo,xHi));
  double wxDstarCh = nBins/(xHi-xLo);
  VxDstarCh[0]->GetXaxis()->SetTitle("x_{E}");
  VxDstarCh[0]->GetYaxis()->SetTitle("1/<n_{D}> dn_{D}/dx");
  VxDstarCh[0]->SetMinimum(0.0);
  VxDstarCh[0]->SetMaximum(2.8);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VxDstarCh.push_back(new TH1D(string("xDstarCh"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleSpectra) {
    vroot.createDisplay("xDstarCh",VxDstarCh,plotSizeX,plotSizeY);
    vroot.setNormalize("xDstarCh",true);
    vroot.setDataFile("xDstarCh",string(VINCIA_DAT+"Aleph-91-xDstarCh.dat").c_str(),20,10,30,40,51,52,61,62);
    vroot.setLogY("xDstarCh",false);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("xDstarCh",mcFileRead[iRead]+"-xDstarCh.dat");
  }

  // x(D*) parents
  nBins = 20;  
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1D*> VmotDstarCh;
  VmotDstarCh.push_back(new TH1D("motDstarCh","x(D^{*#pm}) (x_{E}>0.1)",nBins,xLo,xHi));
  double wmotDstarCh = nBins/(xHi-xLo);
  VmotDstarCh[0]->GetXaxis()->SetTitle("x_{E}");
  VmotDstarCh[0]->GetYaxis()->SetTitle("1/<n_{D}> dn_{D}/dx");
  for (int iMot=1; iMot<=3; iMot++) 
    VmotDstarCh.push_back(new TH1D(string("motDstarCh"+num2str(iMot,0)).c_str(),
                                   " ",nBins,xLo,xHi));
  if (hadronLevel && OPTshowParticleComposition) {
    vroot.createDisplay("motDstarCh",VmotDstarCh[0],plotSizeX,plotSizeY);
    for (int i=1; i<int(VmotDstarCh.size()); ++i) 
      vroot.addTH1DPtr("motDstarCh",VmotDstarCh[i]);
    vroot.setLogY("motDstarCh",false);
    vroot.setNormalize("motDstarCh",false);
    vector<string> legendID;
    legendID.push_back("Total");
    legendID.push_back("Z#rightarrow cc");
    legendID.push_back("Z#rightarrow bb");
    legendID.push_back("Z#rightarrow uu+dd+ss");
    vroot.setLegend("motDstarCh",legendID);
    vroot.setRatioRange("motDstarCh",0.0,1.0);
  }

  // PTIN vs Thrust Axes
  nBins = 140; // Important = multiplum of # of data bins, for ratio rebinning
  xLo =  0.0;
  xHi = 14.0;
  vector<TH1D*> VPTinThr;
  VPTinThr.push_back(new TH1D("PTinThr","Charged p_{T}^{in} (Thrust Axes)",
                              nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VPTinThr[0],rescuePrefix+"PTinThr.dat");
  double wPTinThr = nBins/(xHi-xLo);
  VPTinThr[0]->GetXaxis()->SetTitle("p_{T}^{in}");
  VPTinThr[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dp_{T}");
  VPTinThr[0]->SetMinimum(0.000001);
  VPTinThr[0]->SetMaximum(11.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VPTinThr.push_back(new TH1D(string("PTinThr"+num2str(iVar,0)).c_str(),
                                " ",nBins,xLo,xHi));  
  if (hadronLevel && OPTshowPTinout) {
    vroot.createDisplay("PTinThr",VPTinThr,plotSizeX,plotSizeY);
    vroot.setNormalize("PTinThr",true);
    vroot.setDataFile("PTinThr",string(VINCIA_DAT+"Delphi-91-PTIN.dat").c_str(),10,30,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("PTinThr",mcFileRead[iRead]+"-PTinThr.dat");
    vroot.setLogY("PTinThr");
    vroot.setRebinLikeData("PTinThr");
  }

  // PTIN vs Sphericity Axes
  nBins = 140; // Important = multiplum of # of data bins, for ratio rebinning
  xLo =  0.0;
  xHi = 14.0;
  vector<TH1D*> VPTinSph;
  VPTinSph.push_back(new TH1D("PTinSph","Charged p_{T}^{in} (Sphericity Axes)",
                              nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VPTinSph[0],rescuePrefix+"PTinSph.dat");
  double wPTinSph = nBins/(xHi-xLo);
  VPTinSph[0]->GetXaxis()->SetTitle("p_{T}^{in}");
  VPTinSph[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dp_{T}");
  VPTinSph[0]->SetMinimum(0.000001);
  VPTinSph[0]->SetMaximum(11.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VPTinSph.push_back(new TH1D(string("PTinSph"+num2str(iVar,0)).c_str(),
                                " ",nBins,xLo,xHi));  
  if (hadronLevel && OPTshowPTinout) {
    vroot.createDisplay("PTinSph",VPTinSph,plotSizeX,plotSizeY);
    vroot.setNormalize("PTinSph",true);
    vroot.setDataFile("PTinSph",string(VINCIA_DAT+"Aleph-91-PTIN-s.dat").c_str(),
                      10,30,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("PTinSph",mcFileRead[iRead]+"-PTinSph.dat");
    vroot.setLogY("PTinSph"); 
    vroot.setRebinLikeData("PTinSph");
  }

  // PTOUT vs Thrust Axes
  nBins = 50; // Important = multiplum of # of data bins, for ratio rebinning
  xLo =  0.0;
  xHi =  5.0;
  vector<TH1D*> VPToutThr;
  VPToutThr.push_back(new TH1D("PToutThr","Charged p_{T}^{out} (Thrust Axes)",
                               nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VPToutThr[0],rescuePrefix+"PToutThr.dat");
  double wPToutThr = nBins/(xHi-xLo);
  VPToutThr[0]->GetXaxis()->SetTitle("p_{T}^{out}");
  VPToutThr[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dp_{T}");
  VPToutThr[0]->SetMinimum(0.000001);
  VPToutThr[0]->SetMaximum(11.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VPToutThr.push_back(new TH1D(string("PToutThr"+num2str(iVar,0)).c_str(),
                                 " ",nBins,xLo,xHi));  
  if (hadronLevel && OPTshowPTinout) {
    vroot.createDisplay("PToutThr",VPToutThr,plotSizeX,plotSizeY);
    vroot.setNormalize("PToutThr",true);
    vroot.setDataFile("PToutThr",string(VINCIA_DAT+"Delphi-91-PTOUT.dat").c_str(),
                      10,30,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("PToutThr",mcFileRead[iRead]+"-PToutThr.dat");
    vroot.setLogY("PToutThr");
    vroot.setRebinLikeData("PToutThr");
  }

  // PTOUT vs Sphericity Axes
  nBins = 50; // Important = multiplum of # of data bins, for ratio rebinning
  xLo =  0.0;
  xHi =  5.0;
  vector<TH1D*> VPToutSph;
  VPToutSph.push_back(new TH1D("PToutSph",
    "Charged p_{T}^{out} (Sphericity Axes)",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VPToutSph[0],rescuePrefix+"PToutSph.dat");
  double wPToutSph = nBins/(xHi-xLo);
  VPToutSph[0]->GetXaxis()->SetTitle("p_{T}^{out}");
  VPToutSph[0]->GetYaxis()->SetTitle("1/<n_{ch}> dn_{ch}/dp_{T}");
  VPToutSph[0]->SetMinimum(0.000001);
  VPToutSph[0]->SetMaximum(11.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VPToutSph.push_back(new TH1D(string("PToutSph"+num2str(iVar,0)).c_str(),
                                 " ",nBins,xLo,xHi));  
  if (hadronLevel && OPTshowPTinout) {
    vroot.createDisplay("PToutSph",VPToutSph,plotSizeX,plotSizeY);
    vroot.setNormalize("PToutSph",true);
    vroot.setDataFile("PToutSph",string(VINCIA_DAT+"Aleph-91-PTOUT-s.dat").c_str(),
                      10,30,40,51,52,61,62);
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("PToutSph",mcFileRead[iRead]+"-PToutSph.dat");
    vroot.setLogY("PToutSph");
    vroot.setRebinLikeData("PToutSph");
  }

  // Thrust (non-B flavor)
  nBins = 20; // Important = multiplum of # of data bins, for ratio rebinning
  xLo = 0.0;
  xHi = 0.5;
  vector<TH1D*> VThrustLight;
  VThrustLight.push_back(new TH1D("ThrustLight","1-Thrust (udsc)",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VThrustLight[0],rescuePrefix+"T.dat");
  double wThrust = nBins/(xHi-xLo);
  VThrustLight[0]->GetXaxis()->SetTitle("1-T (udsc)");
  VThrustLight[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/d(1-T)");
  VThrustLight[0]->SetMinimum(0.0001);
  VThrustLight[0]->SetMaximum(1000.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VThrustLight.push_back(new TH1D(string("ThrustLight"+num2str(iVar,0)).c_str()," ",nBins,xLo,xHi));  
  vroot.createDisplay("T",VThrustLight);
  vroot.setNormalize("T",true);
  vroot.setDataFile("T",string(VINCIA_DAT+"L3-91-Thrust-udsc.dat").c_str(),10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    vroot.addHisFile("T",mcFileRead[iRead]+"-T.dat");
  vroot.setLogY("T");
  //vroot.setRatioRange("T",0.75,1.25);

  // C (light)
  nBins = 20; // Important = multiplum of # of data bins, for ratio rebinning
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1D*> VCLight;
  VCLight.push_back(new TH1D("CLight","C Parameter (udsc)",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VCLight[0],rescuePrefix+"C.dat");
  double wC = nBins/(xHi-xLo);
  VCLight[0]->GetXaxis()->SetTitle("C");
  VCLight[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dC");
  VCLight[0]->SetMinimum(0.0005);
  VCLight[0]->SetMaximum(200.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VCLight.push_back(new TH1D(string("CLight"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.createDisplay("C",VCLight,plotSizeX,plotSizeY);
  vroot.setNormalize("C",true);
  vroot.setDataFile("C",string(VINCIA_DAT+"L3-91-C-udsc.dat").c_str(),10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    vroot.addHisFile("C",mcFileRead[iRead]+"-C.dat");
  vroot.setLogY("C");

  // D (udsc)
  nBins = 25; // Important = multiplum of # of data bins, for ratio rebinning
  xLo   = 0.0;
  xHi   = 0.8;
  vector<TH1D*> VDLight;
  VDLight.push_back(new TH1D("DLight","D Parameter (udsc)",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VDLight[0],rescuePrefix+"D.dat");
  double wD = nBins*1.0/(xHi-xLo);
  VDLight[0]->GetXaxis()->SetTitle("D");
  VDLight[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dD");
  VDLight[0]->SetMinimum(0.0005);
  VDLight[0]->SetMaximum(100.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    VDLight.push_back(new TH1D(string("DLight"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.createDisplay("D",VDLight,plotSizeX,plotSizeY);
  vroot.setNormalize("D",true);
  vroot.setDataFile("D",string(VINCIA_DAT+"L3-91-D-udsc.dat").c_str(),10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    vroot.addHisFile("D",mcFileRead[iRead]+"-D.dat");
  vroot.setLogY("D");

  // Wide Jet Broadening
  nBins = 20; // Important = multiplum of # bins in data, for ratio rebinning
  xLo = 0.0;
  xHi = 0.3;
  vector<TH1D*> VBjWide;
  VBjWide.push_back(new TH1D("BWide","Wide Jet Broadening (udsc)",nBins,xLo,xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VBjWide[0],rescuePrefix+"BW.dat");
  double wBjWide = nBins/(xHi-xLo);
  VBjWide[0]->GetXaxis()->SetTitle("B_{W}");
  VBjWide[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dB_{W}");
  VBjWide[0]->SetMaximum(1000.0); 
  VBjWide[0]->SetMinimum(0.001);   
  for (int iVar=1; iVar<nWeights; iVar++) 
    VBjWide.push_back(new TH1D(string("BWide"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.createDisplay("BW",VBjWide,plotSizeX,plotSizeY);
  vroot.setNormalize("BW",true);
  vroot.setDataFile("BW",string(VINCIA_DAT+"L3-91-BWide-udsc.dat").c_str(),20,10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    vroot.addHisFile("BW",mcFileRead[iRead]+"-BW.dat");
  vroot.setLogY("BW");

  // Total Jet Broadening
  nBins = 20; // Must match or be multiplum of # bins in data, for ratio rebinning
  xLo = 0.0;
  xHi = 0.4;
  vector<TH1D*> VBjTot;
  VBjTot.push_back(new TH1D("BTot","Total Jet Broadening (udsc)",
                            nBins, xLo, xHi));
  if (OPTinitialiseFromSaved)
    vroot.initTH1Dptr(VBjTot[0],rescuePrefix+"BT.dat");
  double wBjTot = 200.0/0.4;
  VBjTot[0]->GetXaxis()->SetTitle("B_{T}");
  VBjTot[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dB_{T}");
  VBjTot[0]->SetMaximum(1000.0); 
  VBjTot[0]->SetMinimum(0.001);     
  for (int iVar=1; iVar<nWeights; iVar++) 
    VBjTot.push_back(new TH1D(string("BTot"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.createDisplay("BT",VBjTot,plotSizeX,plotSizeY);
  vroot.setNormalize("BT",true);
  vroot.setDataFile("BT",string(VINCIA_DAT+"L3-91-BTot-udsc.dat").c_str(),20,10,30,40,51,52,61,62);
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    vroot.addHisFile("BT",mcFileRead[iRead]+"-BT.dat");
  vroot.setLogY("BT");

  // Event analysis tools
  Thrust Thrust(1), ThrustCh(3);
  Sphericity SphLin(1,1),SphSqCh(2,3);;

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // Number of quarks, partons, and charged particles
  double weightSum    = 0.0;
  double nQuarksSum   = 0.0;
  double nPartonsSum  = 0.0;
  double nChargedSum  = 0.0;
  double nCharged2sum = 0.0;
  double nAllSumCentral     = 0.0;
  double nChargedSumCentral = 0.0;
  double nMesSum[20] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};
  double nBarSum[12] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};
  double nCharmSum[20] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0. ,0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};
  double nBeautySum[20] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0. ,0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};
  double nEvDch = 0.0;
  double nEvD0 = 0.0;
  
  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    // Generate next event
    if (!pythia.next()) continue;
    eventPtr = &pythia.event;

    // Histogram uncertainty weights
    for (int iVar=0; iVar<nWeights; iVar++) {
      Vweights[iVar]->Fill(log10(max(1.0e-9,vincia.weight(iVar))));
    }
      
    // Count FS charged hadrons, partons, and quarks
    int nCharged = 0;
    int nPartons = 0;
    int nQuarks  = 0;

    // Bools to accept/reject events with specific content
    bool isB       = false;
    bool isDch     = false;
    bool isD0      = false;
    bool hasLambdaC = false;
    bool hasB0s    = false;
    bool hasBplus  = false;
    bool is4b      = false;
    bool hasG2b    = false;
    bool hasG2c    = false;
    bool isBbaryon = false;
    for (int i=1;i<eventPtr->size();i++) {            
      if (abs(eventPtr->at(i).status()) > 70) break;
      int id = eventPtr->at(i).id();
      int iMot1 = eventPtr->at(i).mother1();
      int iMot2 = eventPtr->at(i).mother2();
      int idMot1 = (iMot1 != 0) ? eventPtr->at(iMot1).id() : 0;
      int idMot2 = (iMot2 != 0) ? eventPtr->at(iMot2).id() : 0;
      if ( abs(id) == 5) {
        isB = true;          
        if (i >= 10 && abs(eventPtr->at(i).status()) < 60 
            && abs(eventPtr->at(i).status()) > 30 
            && abs(eventPtr->at(7).id()) == 5
            && (idMot1 != id && idMot2 != id) ) is4b = true;
        if (i >= 10 && abs(eventPtr->at(i).status()) < 60
            && abs(eventPtr->at(i).status()) > 30 
            && (idMot1 == 21 ) ) {
	  hasG2b = true;
	}
      }
      if ( i >= 10 && abs(id) == 4 && abs(eventPtr->at(i).status()) < 60 &&
           abs(eventPtr->at(i).status()) > 30 && (idMot1 != id && idMot2 != id) ) 
        hasG2c = true;        
    }

    // Main loop over event record
    for (int i=1;i<eventPtr->size();i++) {

      // Skip particles that are just copies
      if (eventPtr->at(i).daughter1() == eventPtr->at(i).daughter2() 
          && eventPtr->at(i).daughter1() != 0) continue;

      // Count up final-state charged hadrons
      double x  =  2.0*eventPtr->at(i).pAbs()/eCM;
      double xE =  2.0*eventPtr->at(i).e()/eCM;
      double xi = -log(x);
      int id    =  eventPtr->at(i).id();
      int idp   =  abs(id);
      if (eventPtr->at(i).isCharged() && eventPtr->at(i).status() > 80) {
	nCharged++;
        for (int iVar=0; iVar<nWeights; iVar++) {
          if (vincia.weight(iVar) > 0.0) {
            if (! isB) Vlnx[iVar]->Fill(xi,wLnx*vincia.weight(iVar));
            if (x>0.02) {
	      Vx[iVar]->Fill(x,wx*vincia.weight(iVar));
	    }
	  }
	}
        // Identified-particle composition as function of Lnx
        if (!isB) {
          if (idp == 211) VidLnxSum[0]->Fill( xi, wLnx * vincia.weight(0));
          else if (idp == 321) VidLnxSum[1]->Fill( xi, wLnx * vincia.weight(0));
          else if (idp == 2212) VidLnxSum[2]->Fill( xi, wLnx * vincia.weight(0));
          else VidLnxSum[3]->Fill( xi, wLnx * vincia.weight(0));
        }
      }
      // Find last parton-level partons
      int iDau1 = eventPtr->at(i).daughter1();
      if (iDau1 == 0 || abs((*eventPtr)[iDau1].status()) > 80) {
	// Count up partons and quarks
	if (eventPtr->at(i).isQuark() || eventPtr->at(i).isGluon() ) nPartons++;
	if (eventPtr->at(i).isQuark()) nQuarks++;	
      }
      // Find identified particles
      /* Mesons:
	# 211 0: e+ e- --> pi+- X
	# 111 1: e+ e- --> pi0 X
	# 321 2: e+ e- --> K+- X
	# 221 3: e+ e- --> eta X
	# 331 4: e+ e- --> etaprime(958) X
	# 213 5: e+ e- --> rho(770)+- X
	# 113 6: e+ e- --> rho(770)0 X
	# 223 7: e+ e- --> omega(782) X
	# 323 8: e+ e- --> K*(892)+- X
	# 333 9: e+ e- --> phi(1020) X
	#    10: K* /K- (ee from 10 GeV to 91 GeV + ALICE pp 1208.5717)
        #    11: phi/K* (ee from 10 GeV to 91 GeV + ALICE pp 1208.5717)
        #    12: phi/K- (ee from 10 GeV to 91 GeV + ALICE pp 1208.5717)
        #    13: phi/pi- (ee from 10 GeV to 91 GeV + ALICE pp 1208.5717)
	# 10111 10: e+ e- --> a0(980) X
	# 10221 11: e+ e- --> f0(980) X
      */
      /* Baryons:
	 # 2212      0 : e+ e- --> p (pbar) X
	 # 3122      1: e+ e- --> Lambda (Lambdabar) X
	 # 3122/2212 2: e+ e- --> Lambda (Lambdabar) / p(pbar)
	 # 3122/321  3: e+ e- --> Lambda (Lambdabar) / K+-
	 # 3222+3112 4: e+ e- --> Sigma+- (Sigmabar+-) X
	 # 3212      5: e+ e- --> Sigma0 (Sigmabar0) X
	 # 2224      6 : e+ e- --> Delta(1232)++ (Delta(1232)--) X
	 # 3224+3114 7: e+ e- --> Sigma(1385)+- (Sigmabar(1385)+-) X
	 # 3312      8: e+ e- --> Xi- (Xibar+) X
	 # 3324      9: e+ e- --> Xi(1530)0 (Xibar(1530)0) X
	 # 3334     10: e+ e- --> Omega- (Omegabar+) X 
      */
      /* Charm:
	 # 411    0: e+ e- --> D+- X (HEPDATA/ALEPH)
         # 421    1: e+ e- --> D0 (Dbar0) X (HEPDATA/ALEPH)
         # 413    2: e+ e- --> D*(2010)+- X (PDG S46)
 	 # 431    3: e+ e- --> Ds+- X (PDG S46)
         #        4: e+ e- --> Lambda_c+ X (PDG Z summary)
         #        5: g->cc (OPAL+ALEPH)
         # 443    6: e+ e- --> J/psi X (PDG)
         # 20443  7: e+ e- --> chi(c1) X (PDG)
         # 100443 8: e+ e- --> psi(3685) X (PDG)	
      */
      /* Beauty:
         # 511          0: BR(Z --> B+ X) (PDG Z summary)
         # 511+521      1: e+ e- --> B+-0 X (PDG S46)	
         # 513+523+533  2: e+ e- --> B*+-0 X (PDG S46)	
         # 531          3: BR(Z --> Bs0 X) (PDG Z summary)	
         # 5000 - 5999  4: BR(Z --> B-baryons X) (PDG Z summary)
         #              5: g -> bb
         #              6: BR(Z --> 4b X) (PDG Z summary)
         # 553 + 100553 7: e+e- --> Upsilon(1S+2S+3S) X (PDG Z summary)
      */
      if (idp == 211) nMesSum[0] += vincia.weight(0);
      if (idp == 111) nMesSum[1] += vincia.weight(0);
      if (idp == 321) {
	nMesSum[2] += vincia.weight(0);
        // SLD charged kaons
        /*
	if ( (x >= 0.016 && x <= 0.164) 
	     || (x >= 0.208 && x <= 0.768) ) {
	  for (int iVar=0; iVar<nWeights; iVar++) 
	    VxKchSLD[iVar]->Fill(x,wxKchSLD*vincia.weight(iVar)); 
	}
        */
        // ALEPH, DELPHI, OPAL charged kaons
	if ( (xi <= 5.2 && xi >= 4.0) || (xi <= 3.925) ) {
	  for (int iVar=0; iVar<nWeights; iVar++) {
	    VxKch[iVar]->Fill(-xi,wxKch*vincia.weight(iVar));         
          }
        }
        // Histogram kaon mothers
        int iMot = eventPtr->at(i).mother1();
        int idMot = eventPtr->at(iMot).id();
        VkChLnxSum[0]->Fill( xi, wLnx * vincia.weight(0));
        if (abs(idMot) >= 400 && abs(idMot) < 599) 
          VkChLnxSum[2]->Fill( xi, wLnx * vincia.weight(0));
        else if (abs(idMot) == 313 || abs(idMot) == 323) 
          VkChLnxSum[3]->Fill( xi, wLnx * vincia.weight(0));
        else if (abs(idMot) == 3) 
          VkChLnxSum[1]->Fill( xi, wLnx * vincia.weight(0));
        else if (abs(idMot) == 21 || abs(idMot) <= 2) 
          VkChLnxSum[4]->Fill( xi, wLnx * vincia.weight(0));
        else if (abs(idMot) <= 5)
          VkChLnxSum[5]->Fill( xi, wLnx * vincia.weight(0));
        else VkChLnxSum[6]->Fill( xi, wLnx * vincia.weight(0));
      }
      if (idp == 221) nMesSum[3] += vincia.weight(0);
      if (idp == 331) nMesSum[4] += vincia.weight(0);
      if (idp == 213) nMesSum[5] += vincia.weight(0);
      if (idp == 113) nMesSum[6] += vincia.weight(0);
      if (idp == 223) nMesSum[7] += vincia.weight(0);
      if (idp == 323) nMesSum[8] += vincia.weight(0);
      if (idp == 333) {
        nMesSum[9] += vincia.weight(0);
        if (x > 0.005) {
	  for (int iVar=0; iVar<nWeights; iVar++) {
	    VxPhi[iVar]->Fill(x,wxPhi*vincia.weight(iVar));         
          }
        }
      }
      if (idp == 10111) nMesSum[16] += vincia.weight(0);
      if (idp == 10221) nMesSum[17] += vincia.weight(0);
      if (idp == 2212) {
	nBarSum[0] += vincia.weight(0);
	// ALEPH + SLD combined protons
	if ( (xi <= 4.6047 && xi>=4.012 ) 
	     || (xi <= 3.735 && xi >= 3.578) 
	     || (xi <= 3.4189 && xi >= 1.799) 
	     || (xi <= 1.482) ) 
	  for (int iVar=0; iVar<nWeights; iVar++) {
	    Vxp[iVar]->Fill(-xi,vincia.weight(iVar)*wxp);
          }
      }
      if (idp == 3122) {
	nBarSum[1] += vincia.weight(0);
	if (xi >= 0.4 && xi <= 4.8)
	  for (int iVar=0; iVar<nWeights; iVar++) {
	    VxLam[iVar]->Fill(xi,vincia.weight(iVar)*wxLam);
          }
      }
      if (idp == 3222) nBarSum[2] += vincia.weight(0); 
      if (idp == 3112) nBarSum[2] += vincia.weight(0); 
      if (idp == 3212) nBarSum[3] += vincia.weight(0); 
      if (idp == 3312) {
	nBarSum[4] += vincia.weight(0); 
	if (xE >= 0.03 && xE <= 0.5) 
	  for (int iVar=0; iVar<nWeights; iVar++) {
            // if (iVar != 0 && iVar != 9) continue;
	    VxXi[iVar]->Fill(xE,vincia.weight(iVar)*wxXi);
          }
      }
      if (idp == 2224) nBarSum[5] += vincia.weight(0);
      if (idp == 3114) nBarSum[6] += vincia.weight(0); 
      if (idp == 3224) nBarSum[6] += vincia.weight(0); 
      if (idp == 3324) nBarSum[7] += vincia.weight(0); 
      if (idp == 3334) nBarSum[8] += vincia.weight(0); 

      // Charm
      if (idp == 411) {
        isDch = true;
        nCharmSum[0] += vincia.weight(0);
      }
      if (idp == 421) {
        isD0 = true;
        nCharmSum[1] += vincia.weight(0);
      }
      // D*+-; include spectrum comparison
      if (idp == 413) {
        nCharmSum[2] += vincia.weight(0);
        for (int iVar=0; iVar<nWeights; iVar++) {
          if (xE > 0.1) {
            // if (iVar != 0 && iVar != 9) continue;
            VxDstarCh[iVar]->Fill(xE,vincia.weight(iVar)*wxDstarCh);
          }
        }        
        VmotDstarCh[0]->Fill(xE,wmotDstarCh*vincia.weight(0));
        if (abs(eventPtr->at(6).id()) == 4) 
          VmotDstarCh[1]->Fill(xE,wmotDstarCh*vincia.weight(0));
        else if (abs(eventPtr->at(6).id()) == 5) 
          VmotDstarCh[2]->Fill(xE,wmotDstarCh*vincia.weight(0));
        else 
          VmotDstarCh[3]->Fill(xE,wmotDstarCh*vincia.weight(0));
      }
      if (idp == 431) nCharmSum[3] += vincia.weight(0);
      if (idp == 443) nCharmSum[6] += vincia.weight(0);
      if (id == 4122) hasLambdaC = true;
      if (idp == 20443) nCharmSum[7] += vincia.weight(0);
      if (idp == 100443) nCharmSum[8] += vincia.weight(0);

      // Beauty
      if (id == 521) hasBplus = true;
      if (idp == 511 || idp == 521) nBeautySum[1] += vincia.weight(0);
      if (idp == 513 || idp == 523 || idp == 533)
        nBeautySum[2] += vincia.weight(0);
      if (id  == 531) hasB0s = true; 
      // (Upsilon rate is normalized by factor 10 to fit on plot)
      if (idp == 553 || idp == 100553) nBeautySum[7] += 10.*vincia.weight(0);
      if (idp >= 5000 && idp <= 6000) isBbaryon = true;
    }

    // Fill BR  inclusive branching fractions
    if (hasLambdaC) nCharmSum[4] += vincia.weight(0);
    if (hasG2c) nCharmSum[5] += vincia.weight(0);
    if (hasBplus) nBeautySum[0] += vincia.weight(0);
    if (hasB0s) nBeautySum[3] += vincia.weight(0);
    if (isBbaryon) nBeautySum[4] += vincia.weight(0);
    if (hasG2b) nBeautySum[5] += vincia.weight(0);
    if (is4b) nBeautySum[6] += vincia.weight(0);
    
    // Fill Histograms
    weightSum    += vincia.weight(0);
    nQuarksSum   += 0.5*nQuarks * vincia.weight(0);
    nPartonsSum  += nPartons * vincia.weight(0);
    nChargedSum  += nCharged * vincia.weight(0);
    nCharged2sum += pow2(nCharged) * vincia.weight(0); 
    nEvDch += (isDch ? 1 : 0) * vincia.weight(0);
    nEvD0 += (isD0 ? 1 : 0) * vincia.weight(0);
    HnPartons->Fill(1.0*nPartons,vincia.weight(0));    
    for (int iVar=0; iVar<nWeights; iVar++) {
      if (vincia.weight(iVar) <= 0.0) continue;
      if (! isB) VnCharged[iVar]->Fill(1.0*nCharged,vincia.weight(iVar));
      VnQuarks[iVar]->Fill(0.5*nQuarks,vincia.weight(iVar));
      VnFSR[iVar]->Fill(vincia.fsrShower.getNbranch(1),vincia.weight(iVar));
    }
    
    // Thrust
    Thrust.analyze( *eventPtr );
    ThrustCh.analyze( *eventPtr );
    // Force light quarks only
    if (! isB) {
      for (int iVar=0; iVar<nWeights; iVar++) {
        VThrustLight[iVar]->Fill(1-Thrust.thrust(),wThrust*vincia.weight(iVar));
      }
    }
    
    // Histogram Sphericity values
    if (nCharged >= 2.0) {
      SphLin.analyze( *eventPtr );
      SphSqCh.analyze( *eventPtr );
      double evC=3*(SphLin.eigenValue(1)*SphLin.eigenValue(2)
		    + SphLin.eigenValue(2)*SphLin.eigenValue(3)
		    + SphLin.eigenValue(3)*SphLin.eigenValue(1));
      double evD=27*SphLin.eigenValue(1)*SphLin.eigenValue(2)
	*SphLin.eigenValue(3);
      // Light quarks
      if (! isB) {
        for (int iVar=0; iVar<nWeights; iVar++) {
          VCLight[iVar]->Fill(evC,wC*vincia.weight(iVar));
          VDLight[iVar]->Fill(evD,wD*vincia.weight(iVar));
        }
      }

      // Histogram jet masses
      Vec4 thrustAxis = Thrust.eventAxis(1);   
      Vec4 majorChAxis  = ThrustCh.eventAxis(2);   
      Vec4 minorChAxis  = ThrustCh.eventAxis(3);   
      Vec4 sphAxis1     = SphLin.eventAxis(1);
      //Vec4 sphChAxis    = SphSqCh.eventAxis(1);   
      Vec4 sphChAxis2   = SphSqCh.eventAxis(2);   
      Vec4 sphChAxis3   = SphSqCh.eventAxis(3);   
      Vec4 pPlus, pMinus;
      double bPlus(0.), bMinus(0.), pPlusAbs(0.), pMinusAbs(0.);
      double pToutThr(0.), pTinThr(0.), pToutSph(0.), pTinSph(0.);
      // Loop over event to compute masses
      for (int i=0; i<eventPtr->size(); i++) {
        if (not eventPtr->at(i).isFinal() ) continue;
        int id = eventPtr->at(i).id();
        if (abs(id) == 12 || abs(id) == 14 || abs(id) == 16) continue;
        Particle& pNow = eventPtr->at(i);
        // pTin and pTout
        if (pNow.isCharged() && OPTshowPTinout) {
          pTinThr  = abs(pNow.px()*majorChAxis.px() 
                     + pNow.py()*majorChAxis.py() + pNow.pz()*majorChAxis.pz());
          pToutThr = abs(pNow.px()*minorChAxis.px() 
                     + pNow.py()*minorChAxis.py() + pNow.pz()*minorChAxis.pz());
          pTinSph  = abs(pNow.px()*sphChAxis2.px() 
                     + pNow.py()*sphChAxis2.py() + pNow.pz()*sphChAxis2.pz());
          pToutSph = abs(pNow.px()*sphChAxis3.px() 
                     + pNow.py()*sphChAxis3.py() + pNow.pz()*sphChAxis3.pz());
          for (int iVar=0; iVar<nWeights; iVar++) {
            VPTinThr[iVar]->Fill(pTinThr,wPTinThr*vincia.weight(iVar));
            VPToutThr[iVar]->Fill(pToutThr,wPToutThr*vincia.weight(iVar));
            VPTinSph[iVar]->Fill(pTinSph,wPTinSph*vincia.weight(iVar));
            VPToutSph[iVar]->Fill(pToutSph,wPToutSph*vincia.weight(iVar));
          }          
        }
	// Also count density of all particles in central rapidity unit
	double pzSph =  pNow.px()*sphAxis1.px() 
	  + pNow.py()*sphAxis1.py() + pNow.pz()*sphAxis1.pz();
	double ySph = 0.5 * log( (pNow.e() + pzSph)/(pNow.e() - pzSph) );
	if (abs(ySph) < 0.5) {
	  nAllSumCentral += vincia.weight(0);
	  if (pNow.isCharged()) nChargedSumCentral += vincia.weight(0);
	}
        // Jet masses and broadening: sums
        double dotp = pNow.px()*thrustAxis.px()
          + pNow.py()*thrustAxis.py() + pNow.pz()*thrustAxis.pz();
        double pAbs = pNow.pAbs();
        double cosT = dotp/pAbs/thrustAxis.pAbs();
        double sinT = sqrt(1.-pow2(cosT));
        if (dotp > 0) {
          pPlus     += pNow.p();
          bPlus     += abs(pAbs * sinT);
          pPlusAbs  += pAbs;
        } else {
          pMinus    += pNow.p();
          bMinus    += abs(pAbs * sinT);
          pMinusAbs += pAbs;
        }
        // Charged-particle Rapidity wrt Thrust axis
        if (pNow.isCharged() || pNow.isGluon()) {
          double pz = pNow.px()*thrustAxis.px() 
            + pNow.py()*thrustAxis.py() + pNow.pz()*thrustAxis.pz();
          double E  = pNow.e();
          double yThr = (E-pz != 0.) ? 0.5*log( (E+pz) / (E-pz) ) : 10.0;
          for (int iVar=0; iVar<nWeights; iVar++) {
            VyThr[iVar]->Fill(abs(yThr),wyThr*vincia.weight(iVar));          
          }
        }
      }
      bPlus  /= 2.*(pPlusAbs+pMinusAbs);
      bMinus /= 2.*(pMinusAbs+pPlusAbs);    
      if (! isB) {
        for (int iVar=0; iVar<nWeights; iVar++) {
          VBjWide[iVar]->Fill(max(bPlus,bMinus),
                              wBjWide*vincia.weight(iVar));
          VBjTot[iVar]->Fill(bPlus+bMinus,wBjTot*vincia.weight(iVar));
        }
      }
    }

    // Finalise some histograms before updating displays
    if ( iEvent == vroot.nextUpdate() ) {
      // Update Lnx histograms
      for (int iBin=0; iBin<int(Vlnx[0]->GetNbinsX()); ++iBin) {
        double denom    = Vlnx[0]->GetBinContent(iBin);
        double denomErr = Vlnx[0]->GetBinError(iBin);
        for (int i=0; i<int(VidLnx.size()); ++i) {        
          double num    = VidLnxSum[i]->GetBinContent(iBin);
          double numErr = VidLnxSum[i]->GetBinError(iBin);
          if (num > 0. && denom >0.) {
            double val    = num/denom;
            double errRel = sqrt(pow2(denomErr/denom)+pow2(numErr/num));
            double err    = errRel * val;
            VidLnx[i]->SetBinContent(iBin,val);
            VidLnx[i]->SetBinError(iBin,err);
          } else {
            //VidLnx[i]->SetBinContent(iBin,0.5);
            //VidLnx[i]->SetBinError(iBin,0.5);
          }
        }
      }
      for (int iBin=0; iBin<int(VkChLnx[0]->GetNbinsX()); ++iBin) {
        for (int i=0; i<int(VkChLnx.size()); ++i) {        
          double num    = VkChLnxSum[i]->GetBinContent(iBin);
          double numErr = VkChLnxSum[i]->GetBinError(iBin);
          VkChLnx[i]->SetBinContent(iBin,num);
          VkChLnx[i]->SetBinError(iBin,numErr);
        }
      }
      // Update PID histograms
      if (nChargedSum > 0) {
        VMesons[0]->Reset();
        // Meson rates / <nCh>
        for (int iMes=0; iMes<=9; iMes++) {
	  // Normalise charged particles by extra factor 1/2 to get per-charge
	  double fac = 1.;
	  if (iMes == 0 || iMes == 2 || iMes == 5 || iMes == 8) fac = 0.5;
          VMesons[0]->Fill(double(iMes),fac*nMesSum[iMes]/nChargedSum);
          double err = fac * sqrt(nMesSum[iMes]) / nChargedSum;
          VMesons[0]->SetBinError(double(iMes)+1,err);
        }
        // Meson ratios
        VMesons[0]->Fill(10.,nMesSum[8]/nMesSum[2]);
        VMesons[0]->SetBinError(11.,sqrt(nMesSum[8])/nMesSum[2]);
        VMesons[0]->Fill(11.,2.*nMesSum[9]/nMesSum[8]);
        VMesons[0]->SetBinError(12.,2.*sqrt(nMesSum[9])/nMesSum[7]);
        VMesons[0]->Fill(12.,2.*nMesSum[9]/nMesSum[2]);
        VMesons[0]->SetBinError(13.,2.*sqrt(nMesSum[9])/nMesSum[2]);
	double sumRho = nMesSum[5]+nMesSum[6];
        VMesons[0]->Fill(13.,3*nMesSum[9]/sumRho);
        VMesons[0]->SetBinError(14.,3.*sqrt(nMesSum[9])/sumRho);
	VMesons[0]->Fill(14.,nMesSum[9]/nMesSum[7]);
	VMesons[0]->SetBinError(15.,sqrt(nMesSum[9])/nMesSum[7]);
        VMesons[0]->Fill(15.,2.*nMesSum[9]/nMesSum[0]);
        VMesons[0]->SetBinError(16.,2.*sqrt(nMesSum[9])/nMesSum[0]);
        // Baryon rates / <nCh>
	// Normalise by 1/2 to get per (anti)baryon
        VBaryons[0]->Reset();
        for (int iBar=0; iBar<=10; iBar++) {
	  double fac = 0.5;
	  if (iBar == 9 || iBar == 10) fac = 1.0;
	  else if (iBar == 2 || iBar == 6) fac = 0.25;
          VBaryons[0]->Fill(double(iBar),fac*nBarSum[iBar]/nChargedSum);
          double err = fac*sqrt(nBarSum[iBar]) / nChargedSum;
          VBaryons[0]->SetBinError(double(iBar)+1,err);
        }
        // Baryon ratios: fill Lambda/p and Lambda/K
        VBaryons[0]->Fill(double(9),nBarSum[1]/nBarSum[0]);
        double err = sqrt(nBarSum[1]) / nBarSum[0];
        VBaryons[0]->SetBinError(double(9)+1,err);
        VBaryons[0]->Fill(double(10),nBarSum[1]/nMesSum[2]);
        err = sqrt(nBarSum[1]) / nMesSum[2];
        VBaryons[0]->SetBinError(double(10)+1,err);
        // Charm
        VCharm[0]->Reset();
        for (int iMes=0; iMes<=19; iMes++) {
          VCharm[0]->Fill(double(iMes),nCharmSum[iMes]/weightSum);
          err = sqrt(nCharmSum[iMes])/weightSum;
          VCharm[0]->SetBinError(double(iMes)+1,err);
        }
        // Beauty
        VBeauty[0]->Reset();
        for (int iMes=0; iMes<=19; iMes++) {
          VBeauty[0]->Fill(double(iMes),nBeautySum[iMes]/weightSum);
          err = sqrt(nBeautySum[iMes])/weightSum;
          // Note: Upsilon rate was artificially multiplied by 10
          if (iMes == 7) err *= sqrt(10.);
          VBeauty[0]->SetBinError(double(iMes)+1,err);
        }
      }
    }
    
    // Call VinciaRoot::update(). Write out extra info when hists were saved.
    vroot.update(); 
    if ( iEvent >= iSave ) {
      iSave       = vroot.nextSave();
      double rmsd = sqrt(nCharged2sum/weightSum-pow2(nChargedSum/weightSum));
      double rmsdn = rmsd/(nChargedSum/weightSum);
      cout<<" At event = "<<iEvent<<"   <nCh> = "<<nChargedSum/weightSum
	  <<" RMSD/<nCh> = "<<rmsdn
	  <<" <nPartons> = "<<nPartonsSum/weightSum
	  <<" <dN(ch)/dy>|y=0 = "<<nAllSumCentral/weightSum
	  <<" ("<<nChargedSumCentral/weightSum<<")"
	  <<"  (next save at "<<iSave<<")"<<endl;
      
    }
    
  }

  // End-of-run forced display update and saves
  vroot.update(vroot.nextSave());
  
  // Print out Vincia internal histograms
  if (vinciaOn && settingsPtr->mode("Vincia:verbose") >= 2) vincia.printHistos();
  
  // Print out Vincia end-of-run information
  pythia.stat();
  if (vinciaOn) vincia.printInfo();

  cout << endl << "Enter q to terminate (will kill displays)..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
