// vincia05-root is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to run Z->bbbar at LEP 1, 
// interfaced to ROOT, including a runtime updated display,
// using VINCIA's ROOT-based runtime display. 
// Uses VINCIA for showering and matching and PYTHIA 8 for all else
// Copyright (c) 2017 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia/Vincia.h"
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot/VinciaRoot.h"
// FASTJET
// #include "fastjet/PseudoJet.hh"
// #include "fastjet/ClusterSequence.hh"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects
  Pythia pythia; 
  VinciaPlugin vincia(&pythia);

  // Read user settings from file
  pythia.readFile("vincia05.cmnd");

   // Global flag whether to exclude neutrinos from clusterings, E_vis, etc
  // (Not yet sure exactly what was done in the L3 analyses)
  bool excludeNeutrinos = false;

  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Shorthand for event record
  Event& event = pythia.event;
  Settings* settingsPtr = vincia.settingsPtr;        // to settings database

  // Extract settings to be used in the main program.  
  double eCM         = settingsPtr->parm("Beams:eCM");
  int    nEvent      = settingsPtr->mode("Main:numberOfEvents");
  bool   vinciaOn    = settingsPtr->flag("Vincia");
  bool   hadronLevel = settingsPtr->flag("HadronLevel:all");
  int    verbose     = settingsPtr->mode("Vincia:verbose");

  // Data path directory, extract from VINCIADATA
  string VINCIA_SHARE     = vincia.VINCIA_SHARE;
  string VINCIA_VAL       = VINCIA_SHARE+"validation/";
  string VINCIA_DAT       = VINCIA_SHARE+"data/";

  //************************************************************************
  // VinciaRoot
  //************************************************************************

  // Create and initialise VinciaRoot object 
  VinciaRoot vroot(settingsPtr);
  vroot.init();

  // Set default plot properties
  vroot.setLogY();                         // Logarithmic Y axes
  vroot.setRatioRange("all",0.5,1.5);      // Range on ratio plots (default is [0.5,1.5])
  vroot.setShowLegendChi2("all",false);    // Compute and display chi2(5%) values?
  vroot.setLeftLabel("all","ee#rightarrowhadrons"); // Left-hand top label
  vroot.setRightLabel("all","91.2 GeV");            // Right-hand top label
  
  // Tell VinciaRoot which verbose level to use
  vroot.setVerbose("all",verbose);
  vroot.setShowLegendChi2("all",false);  

  // Set label for this run
  string mcLabel = (vinciaOn) ? "Vincia" : "Pythia";
  if (!hadronLevel) mcLabel += " partons";
  // Set VinciaRoot plot labels, legends, etc
  vroot.setLegend("all",mcLabel);

  // Compare current run to files with the following prefix(es)
  vector<string> mcFileRead;
  mcFileRead.push_back(string(VINCIA_VAL+"vincia2/vincia05").c_str());
  mcFileRead.push_back(string(VINCIA_VAL+"pythia/vincia05").c_str());
  //mcFileRead.push_back(string(VINCIA_VAL+"vincia1/vincia05").c_str());

  // Next event at which VinciaRoot will save displays/histograms
  int iSave = vroot.nextSave();

  //************************************************************************
  // Initialize VINCIA (automatically initializes PYTHIA) 
  //************************************************************************

  // Initialize
  vincia.init();

  //***************************************************************************
  // ROOT
  //***************************************************************************

  // Form name for ROOT file
  ostringstream outfile;
  // Put root files in subdirectory defined at beginning of file
  if (not vinciaOn) {
    outfile<<"vincia05-pythia";
  } else {
    outfile<<"vincia05";
  }
  if (hadronLevel) outfile<<"-had";
  outfile<<".root";

  //***************************************************************************
  // Book Histograms
  
  // Vincia uncertainty bands: number of variations expected 
  int nWeights = max(1,vincia.nWeights());

  // Default plot sizes
  double plotSizeX = 360;
  double plotSizeY = 420;

  // Thrust
  int nBins = 36;
  double xLo = 0.0;
  double xHi = 0.45;
  vector<TH1D*> HThrust;
  HThrust.push_back(new TH1D("Thrust","1-T (b)", nBins, xLo, xHi));
  double wThrust = nBins/(xHi-xLo);
  HThrust[0]->GetXaxis()->SetTitle("1-T (b)");
  HThrust[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/d(1-T)");
  HThrust[0]->SetMinimum(0.0001);
  HThrust[0]->SetMaximum(500.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    HThrust.push_back(new TH1D(string("Thrust"+num2str(iVar,1)).c_str()," ",nBins,xLo,xHi));
  vroot.createDisplay("T",HThrust,plotSizeX,plotSizeY);
  vroot.setDataFile("T",string(VINCIA_DAT+"L3-91-Thrust-b.dat"),12,32,40,51,52,61,62);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    vroot.addHisFile("T",mcFileRead[iFile]+"-T.dat");
  vroot.setLogY("T");
  vroot.setNormalise("T");
  
  // C
  nBins = 40;
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1D*> HC;  
  HC.push_back(new TH1D("C","C parameter (b)", nBins, xLo, xHi));
  double wC = nBins/(xHi-xLo);
  HC[0]->GetXaxis()->SetTitle("C (b)");
  HC[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dC");
  HC[0]->SetMinimum(0.0005);
  HC[0]->SetMaximum(200.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    HC.push_back(new TH1D(string("C"+num2str(iVar,1)).c_str()," ",nBins,xLo,xHi));
  vroot.createDisplay("C",HC,plotSizeX,plotSizeY);
  vroot.setDataFile("C",string(VINCIA_DAT+"L3-91-C-b.dat"),10,30,40,51,52,61,62);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    vroot.addHisFile("C",mcFileRead[iFile]+"-C.dat");
  vroot.setNormalise("C");
  
  // D (4-jet obs)
  nBins = 50;
  xLo   = 0.0;
  xHi   = 0.8;
  vector<TH1D*> HD;
  HD.push_back(new TH1D("D","D parameter (b)", nBins, xLo, xHi));
  double wD = nBins/(xHi-xLo);
  HD[0]->GetXaxis()->SetTitle("D (b)");
  HD[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dD");
  HD[0]->SetMinimum(0.0005);
  HD[0]->SetMaximum(100.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    HD.push_back(new TH1D(string("D"+num2str(iVar,1)).c_str()," ",nBins,xLo,xHi));
  vroot.createDisplay("D",HD,plotSizeX,plotSizeY);
  vroot.setDataFile("D",string(VINCIA_DAT+"L3-91-D-b.dat"),10,30,40,51,52,61,62);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    vroot.addHisFile("D",mcFileRead[iFile]+"-D.dat");
  vroot.setNormalise("D");
  
  // Heavy-Jet Mass (3-jet obs)
  nBins = 40;
  xLo   = 0.0;
  xHi   = 0.3;
  vector<TH1D*> HMjH;
  HMjH.push_back(new TH1D("HeavyJetMass","Heavy Jet Mass (b)",nBins, xLo, xHi));
  double wMjH = 40.0/0.3;
  HMjH[0]->SetMaximum(200.01); 
  HMjH[0]->SetMinimum(0.01);
  for (int iVar=1; iVar<nWeights; iVar++) 
    HMjH.push_back(new TH1D(string("HeavyJetMass"+num2str(iVar,1)).c_str()," ",nBins,xLo,xHi));
  vroot.createDisplay("MH",HMjH,plotSizeX,plotSizeY);
  vroot.setDataFile("MH",string(VINCIA_DAT+"L3-91-MH-b.dat"),10,30,40,51,52,61,62);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    vroot.addHisFile("MH",mcFileRead[iFile]+"-MH.dat");
  vroot.setNormalise("MH");

  // Wide Jet Broadening
  nBins = 40;
  xLo   = 0.0;
  xHi   = 0.3;
  vector<TH1D*> HBjWide;
  HBjWide.push_back(new TH1D("WideJetBroadening","Wide Jet Broadening (b)",nBins,xLo,xHi));
  double wBjWide = nBins/(xHi-xLo);
  HBjWide[0]->GetXaxis()->SetTitle("B_{W} (b)");
  HBjWide[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dB_{W}");
  HBjWide[0]->SetMaximum(1000.01); 
  HBjWide[0]->SetMinimum(0.001);
  for (int iVar=1; iVar<nWeights; iVar++) 
    HBjWide.push_back(new TH1D(string("WideJetBroadening"+num2str(iVar,1)).c_str()," ",nBins,xLo,xHi));
  vroot.createDisplay("BW",HBjWide,plotSizeX,plotSizeY);
  vroot.setDataFile("BW",string(VINCIA_DAT+"L3-91-BWide-b.dat"),10,30,40,51,52,61,62);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    vroot.addHisFile("BW",mcFileRead[iFile]+"-BW.dat");
  vroot.setNormalise("BW");

  // Total Jet Broadening
  nBins = 40;
  xLo   = 0.0;
  xHi   = 0.4;
  vector<TH1D*> HBjTot;
  HBjTot.push_back(new TH1D("TotalJetBroadening","Total Jet Broadening (b)",nBins, xLo, xHi));
  double wBjTot = nBins/(xHi-xLo);
  HBjTot[0]->GetXaxis()->SetTitle("B_{T} (b)");
  HBjTot[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dB_{T}");
  HBjTot[0]->SetMaximum(1000.); 
  HBjTot[0]->SetMinimum(0.001);
  for (int iVar=1; iVar<nWeights; iVar++) 
    HBjTot.push_back(new TH1D(string("TotalJetBroadening"+num2str(iVar,1)).c_str()," ",nBins,xLo,xHi));
  vroot.createDisplay("BT",HBjTot,plotSizeX,plotSizeY);
  vroot.setDataFile("BT",string(VINCIA_DAT+"L3-91-BTot-b.dat"),10,30,40,51,52,61,62);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    vroot.addHisFile("BT",mcFileRead[iFile]+"-BT.dat");
  vroot.setNormalise("BT");

  // nCharged
  nBins = 60;
  xLo   = -0.5;
  xHi   = 59.5;
  vector<TH1D*> HnCharged;
  HnCharged.push_back(new TH1D("nCharged","Charged Multiplicity (b)", 
                               nBins, xLo, xHi));
  HnCharged[0]->GetXaxis()->SetTitle("n_{ch} (b)");
  HnCharged[0]->GetYaxis()->SetTitle("Probability(n_{ch})");
  HnCharged[0]->SetMinimum(0.000001);
  HnCharged[0]->SetMaximum(100.0);
  for (int iVar=1; iVar<nWeights; iVar++) 
    HnCharged.push_back(new TH1D(string("nCharged"+num2str(iVar,1)).c_str()," ",nBins,xLo,xHi));
  vroot.createDisplay("Nch",HnCharged,plotSizeX,plotSizeY);
  vroot.setDataFile("Nch",string(VINCIA_DAT+"L3-91-Nch-b.dat"),20,40,50,60);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    vroot.addHisFile("Nch",mcFileRead[iFile]+"-Nch.dat");
  vroot.setNormalise("Nch");

  // Ln(1/x)
  nBins = 200;
  xLo   = 0.0;
  xHi   = 8.0;
  vector<TH1D*> Hx;
  Hx.push_back(new TH1D("LnXInv","Charged Momentum Fraction (b)", nBins, xLo, xHi));
  double wx = nBins/(xHi-xLo);
  Hx[0]->GetXaxis()->SetTitle("|Ln(x_{p})| (b)");
  Hx[0]->GetYaxis()->SetTitle("1/n_{ch} dn_{ch}/d|Ln(x_{p})|");
  Hx[0]->SetMinimum(0.00001);
  Hx[0]->SetMaximum(500.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    Hx.push_back(new TH1D(string("LnXInv"+num2str(iVar,1)).c_str()," ",nBins,xLo,xHi));
  vroot.createDisplay("x",Hx,plotSizeX,plotSizeY);
  vroot.setDataFile("x",string(VINCIA_DAT+"L3-91-LogInvx-b.dat"),10,30,40,51,52,61,62);
  vroot.setRebinLikeData("x");
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    vroot.addHisFile("x",mcFileRead[iFile]+"-x.dat");
  vroot.setNormalise("x");

  // xB
  nBins = 50;
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1D*> HxBSLD;
  HxBSLD.push_back(new TH1D("XBSLD","x_{B}^{weak} (SLD)", nBins, xLo, xHi));
  double wxBSLD = nBins/(xHi-xLo);
  HxBSLD[0]->GetXaxis()->SetTitle("x");
  HxBSLD[0]->GetYaxis()->SetTitle("1/N dN/dx_{B}");
  HxBSLD[0]->SetMinimum(0.02);
  HxBSLD[0]->SetMaximum(20.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    HxBSLD.push_back(new TH1D(string("XBSLD"+num2str(iVar,1)).c_str()," ",nBins,xLo,xHi));
  vroot.createDisplay("xB-SLD",HxBSLD,plotSizeX,plotSizeY);
  vroot.setDataFile("xB-SLD",string(VINCIA_DAT+"SLD-91-xBweak.dat"),20,10,30,40,51,52,61,62,71,72);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    vroot.addHisFile("xB-SLD",mcFileRead[iFile]+"-xB-SLD.dat");
  vroot.setNormalise("xB-SLD");

  nBins = 50;
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1D*> HxBDelphi;
  HxBDelphi.push_back(new TH1D("XBDelphi","x_{B}^{weak} (DELPHI)", nBins, xLo, xHi));
  double wxBDelphi = nBins/(xHi-xLo);
  HxBDelphi[0]->GetXaxis()->SetTitle("x");
  HxBDelphi[0]->GetYaxis()->SetTitle("1/N dN/dx_{B}");
  HxBDelphi[0]->SetMinimum(0.02);
  HxBDelphi[0]->SetMaximum(20.);
  for (int iVar=1; iVar<nWeights; iVar++) 
    HxBDelphi.push_back(new TH1D(string("XBDelphi"+num2str(iVar,1)).c_str()," ",nBins,xLo,xHi));
  vroot.createDisplay("xB-Delphi",HxBDelphi,plotSizeX,plotSizeY);
  vroot.setDataFile("xB-Delphi",string(VINCIA_DAT+"Delphi-91-xBweak.dat"),10,30,40,50,60,70);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    vroot.addHisFile("xB-Delphi",mcFileRead[iFile]+"-xB-Delphi.dat");
  vroot.setNormalise("xB-Delphi");

  nBins = 41;
  xLo   = -0.5;
  xHi   = 40.5;
  vector<TH1D*> HxBmoments;
  HxBmoments.push_back(new TH1D("XBmoments","x_{B}^{weak} (moments)", nBins, xLo, xHi)); 
  HxBmoments[0]->GetXaxis()->SetTitle("N");
  HxBmoments[0]->GetYaxis()->SetTitle("Moment");
  HxBmoments[0]->SetMinimum(-0.18);
  HxBmoments[0]->SetMaximum(0.9);
  vroot.createDisplay("xB-moments",HxBmoments,plotSizeX,plotSizeY);
  vroot.setDataFile("xB-moments",string(VINCIA_DAT+"LEP-91-xBmoments.dat"),20,40,50);
  //vroot.setLinY("xB-moments");
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    vroot.addHisFile("xB-moments",mcFileRead[iFile]+"-xB-moments.dat");
  vroot.setNormalize("xB-moments",false);

  // Event analysis tools with or without neutrinos included
  int select = 1;
  if (excludeNeutrinos) select = 2;
  Thrust Thrust(select);
  Sphericity SphLin(1,select),SphSq(2,select);
  ClusterJet ktClu("Durham",select,2);

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // (Optionally) re-initialize random number engine before event loop
  const int initseed = settingsPtr->mode("Random:seed");
  vincia.rndmPtr->init(initseed);

  // Initialize counters
  double nHasB = 0.0;
  double xBsumPows[42];
  for (int iPow = 0; iPow < 41; ++iPow) xBsumPows[iPow] = 0.;  

  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    if (!pythia.next()) continue;

    // Count FS charged hadrons, partons, and quarks
    int nc=0;
    int np=0;
    int nq=0;

    // Check for presence of b quark and compute visible energy 
    Vec4 pVis(0.0);
    bool hasB = false;
    for (int i=1;i<event.size();i++) {

      if (pythia.event[i].id() == 5) hasB = true;

      // Final state only
      if (!pythia.event[i].isFinal())        continue;
      
      // No neutrinos
      if ( excludeNeutrinos ) 
        if (pythia.event[i].idAbs() == 12 || pythia.event[i].idAbs() == 14 ||
            pythia.event[i].idAbs() == 16)     continue;
      
      pVis += event[i].p();
    }
    // Skip event if no b quarks
    if ( !hasB ) continue;
    else ++nHasB;

    // Energy component
    double eVis = pVis.e();
                                                
    for (int i=1;i<event.size();i++) {
      // Count up final-state charged hadrons
      if (event[i].isCharged() && event[i].isFinal()) {
        nc++;
        double LnXInv=log(double(1.0/(2.0*event[i].pAbs()/eCM)));
        for (int iVar=0; iVar<nWeights; iVar++) {
          Hx[iVar]->Fill(LnXInv, wx*vincia.weight(iVar));
        }
      }
      // Find last parton-level partons
      int iDau1 = event[i].daughter1();
      if (iDau1 == 0 || abs(event[iDau1].status()) > 80) {
        // Count up partons and quarks
        if (event[i].isQuark() || event[i].isGluon() ) np++;
        if (event[i].isQuark()) nq++;        
      }
      double x = 2.*event[i].pAbs() / eCM;      

      // Find weakly decaying B mesons 
      int idp = abs(event[i].id());
      double xBweak = -1.0;
      if (hadronLevel) {
        if (idp > 510 && idp < 600) {
          if (abs(event[event[i].daughter1()].id()) < 500) {
            xBweak = x;
          }
        }
        // Find weakly decaying B baryons
        if (idp > 5100 && idp < 6000) {
          if (abs(event[event[i].daughter1()].id()) < 5000) {
            xBweak = x;
          }        
        }
      } 
      // If hadronization switched off, just plot x of b quarks
      else {
        if (idp == 5 && event[i].isFinal()) xBweak = x;
      }
      if (xBweak > 0.) {
        for (int iVar=0; iVar<nWeights; iVar++) {
          if (xBweak > 0.12)
            HxBSLD[iVar]->Fill(xBweak, wxBSLD*vincia.weight(iVar));
          if (xBweak > 0.10)
            HxBDelphi[iVar]->Fill(xBweak, wxBDelphi*vincia.weight(iVar));
        }
        ++xBsumPows[0];
        for (int iPow=1; iPow < 41; ++iPow) 
          xBsumPows[iPow] += pow(xBweak,iPow);
      }
    }

    // Fill charged-particle histogram
    for (int iVar=0; iVar<nWeights; iVar++) {
      HnCharged[iVar]->Fill(1.0*nc, vincia.weight(iVar));
    }
    
    // Event Shape analysis
    Thrust.analyze( event );
    for (int iVar=0; iVar < nWeights; iVar++) {
      HThrust[iVar]->Fill(1-Thrust.thrust(),wThrust*vincia.weight(iVar));
    }

    // Histogram Linear Sphericity values
    if (np >= 2.0) {
      SphLin.analyze( event );
      SphSq.analyze( event );
      double evC=3*(SphLin.eigenValue(1)*SphLin.eigenValue(2)
                    + SphLin.eigenValue(2)*SphLin.eigenValue(3)
                    + SphLin.eigenValue(3)*SphLin.eigenValue(1));
      double evD=27*SphLin.eigenValue(1)*SphLin.eigenValue(2)
        *SphLin.eigenValue(3);
      for (int iVar=0; iVar < nWeights; iVar++) {
        HC[iVar]->Fill(evC,wC*vincia.weight(iVar));
        HD[iVar]->Fill(evD,wD*vincia.weight(iVar));
      }
      
    }

    // Histogram jet masses
    Vec4 thrustAxis = Thrust.eventAxis(1);   
    Vec4 pPlus, pMinus;
    double bPlus(0.), bMinus(0.), pPlusAbs(0.), pMinusAbs(0.);
    // Loop over event to compute masses
    for (int i=0; i<event.size(); i++) {
      if (not event[i].isFinal() ) continue;
      int id = event[i].id();
      if ( excludeNeutrinos )
        if (abs(id) == 12 || abs(id) == 14 || abs(id) == 16) continue;
      double dotp = event[i].px()*thrustAxis.px()
        + event[i].py()*thrustAxis.py()
        + event[i].pz()*thrustAxis.pz();
      double pAbs = event[i].pAbs();
      double cosT = dotp/pAbs/thrustAxis.pAbs();
      double sinT = sqrt(1.-pow2(cosT));
      if (dotp > 0) {
        pPlus     += event[i].p();
        bPlus     += abs(pAbs * sinT);
        pPlusAbs  += pAbs;
      } else {
        pMinus += event[i].p();
        bMinus    += abs(pAbs * sinT);
        pMinusAbs += pAbs;
      }
    }
    bPlus  /= 2.*(pPlusAbs+pMinusAbs);
    bMinus /= 2.*(pMinusAbs+pPlusAbs);    
    double mPlus  = pPlus.mCalc();
    double mMinus = pMinus.mCalc();
    for (int iVar=0; iVar < nWeights; iVar++) {
      HMjH[iVar]->Fill(pow2(max(mPlus,mMinus)/eVis),wMjH*vincia.weight(iVar));
      HBjWide[iVar]->Fill(max(bPlus,bMinus),wBjWide*vincia.weight(iVar));
      HBjTot[iVar]->Fill(bPlus+bMinus,wBjTot*vincia.weight(iVar)); 
    }  
    
    // Compute moments of b-fragmentation function
    if (xBsumPows[0] >= 1) {
      for (int iPow = 1; iPow < 40; ++iPow) {
        double momPow = xBsumPows[iPow]/xBsumPows[0];
        double err    = 1./sqrt(xBsumPows[0]) * sqrt(momPow);
        HxBmoments[0]->SetBinContent(iPow+2,momPow);
        HxBmoments[0]->SetBinError(iPow+2,err);
      }
    }

    // Check if displays should be updated (and/or histograms saved)
    // Pass iEvent to update, since we are skipping quite a few events
    vroot.update(iEvent); 
    if ( iEvent >= iSave ) {
      iSave       = vroot.nextSave();
      cout<<" At event = "<< iEvent <<"  (next save at "<<iSave<<")"<<endl;      
    }
  }
  
  // End-of-run forced display update and saves
  vroot.update(vroot.nextSave());

  // Print out Vincia internal histograms
  if (vinciaOn && verbose >= 2) vincia.printHistos();

  // Print out Vincia end-of-run information
  pythia.stat();
  if (vinciaOn) vincia.printInfo();

  // Save histograms in ROOT file
  cout<<"Saving root file... "<<outfile.str()<<endl;
  // Open ROOT file
  TFile f(outfile.str().c_str(),"recreate");
  for (int iVar=0; iVar<nWeights; iVar++) {
    HThrust[iVar]->Write();
    HC[iVar]->Write();
    HD[iVar]->Write();
    HMjH[iVar]->Write();
    HBjWide[iVar]->Write();
    HBjTot[iVar]->Write();
    HnCharged[iVar]->Write();
    Hx[iVar]->Write();
    HxBSLD[iVar]->Write();
    HxBDelphi[iVar]->Write();
  }
  f.Close();

  cout << endl << "Enter q to terminate (will kill displays)..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
