// AntennaFunctions.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2017 P. Skands
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the AntennaFSR and AntennaISR
// classes, their derived classes (implementations of each specific ISR and
// FSR antenna function), and the AntennaSetFSR and AntennaSetISR classes.
  
#ifndef VINCIAANTENNAFUNCTIONS_H
#define VINCIAANTENNAFUNCTIONS_H

#include "Pythia8/Basics.h"
#include "Pythia8/Event.h"
#include "Pythia8/PythiaStdlib.h"
#include "Vincia/VinciaCommon.h"      
#include "Vincia/SubtractionTerms.h"   // needed for DGLAP tests

namespace Vincia {

//==========================================================================

// The AntennaFSR class:
// Base class implementation for Final-Final Antenna objects

class AntennaFSR {
  
 public:
  
  // Constructor
  AntennaFSR() {
    // Init flags
    isInitPtr = false;
    isInit    = false;      
    // Default values (reset by init)
    verbose              = 1;
    // Initialize LH and RH maps (DO NOT CHANGE)
    RH[9]  = true;
    RH[1]  = true;
    RH[-1] = false;
    LH[9]  = true;
    LH[1]  = false;
    LH[-1] = true;
    chargeFacSav   = 0.0;
    kineMapTypeSav = 0;
    alphaSav       = 0.0;
    modeSLC        = -1;
  };
    
  // Destructor 
  virtual ~AntennaFSR() {};
    
  //----------------------------------------------------------------------
  // List of VIRTUAL functions to be redefined by inheriting classes

  // Names of this antenna, for VINCIA, and for humans
  virtual string vinciaName()   const {return "Vincia:Antenna";};

  // Parton types (AB -> IJK): needed by soft- and collinear-limit checks 
  virtual int idA()    const {return 0;}
  virtual int idB()    const {return 0;}
  virtual int idj()    const {return 0;}

  // The (dimensionless) antenna function.
  // Most generic form including masses and helicities.
  // Default parameter assignments => massless and unpolarized.
  virtual double antennaFunction(double yij, double yjk, 
    double mui=0.0, double muj=0.0, double muk=0.0,
    int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);
    
  // Optional implementation of the AP kernels corresponding
  // to the two sides of the antenna, for collinear-limit checks
  // Defined as PI/yij + PK/yjk, i.e., equivalent to antennae.
  virtual double AltarelliParisi(double yij, double yjk,
    double mui=0., double muj=0., double muk=0.,
    int hA=9, int hB=9, int hI=9, int hJ=9, int hK=9);
  
  //----------------------------------------------------------------------
  // List of VIRTUAL functions that MAY be redefined (if needed)

  // Default initialization 
  virtual bool init();
      
  // Construct baseName from idA, idB, and idj
  virtual string baseName()    const {
    string name = id2str(idj()) + "/" + id2str(idA()) + id2str(idB());
    return name;
  }

  // Wrapper that can modify baseName to more human readable form if required.
  virtual string humanName()   const {
    return baseName();
  }

  // Function to check singularities, positivity, etc.
  virtual bool check();

  //----------------------------------------------------------------------
  // List of NON-VIRTUAL functions (i.e., that may not be redefined)

  // Initialize pointers:
  //    ParticleData : required for masses, color charges, etc
  //    Rndm : required to generate random test points to check antenna
  //    Settings : required for all coefficients, switches, etc.
  void initPtr(Functions* functionsPtrIn, ParticleData* particleDataPtrIn, 
    Rndm* rndmPtrIn, Settings* settingsPtrIn) {
    functionsPtr    = functionsPtrIn;
    particleDataPtr = particleDataPtrIn; 
    rndmPtr         = rndmPtrIn;
    settingsPtr     = settingsPtrIn; 
    isInitPtr       = true;
    isInit          = false;
  };

  // -------
  // Functions to GET

  // Shorthand to get the massless antenna function
  double antennaMassless(double yij, double yjk, int hA=9, int hB=9, 
    int hi=9, int hj=9, int hk=9) {
    return antennaFunction(yij,yjk,0.0,0.0,0.0,hA,hB,hi,hj,hk);
  }

  // Get parameters
  double getChargeFac() { return chargeFacSav; }
  int    getKineMapType() { return kineMapTypeSav; }
  double getAlpha() { return alphaSav; }

  // Functions to get Altarelli-Parisi energy fractions from scaled invariants
  double zi(double yij,double yjk) {return (1.-yjk)/(1.+yij);}
  double zk(double yij,double yjk) {return (1.-yij)/(1.+yjk);}

  //----------------------------------------------------------------------
  // id2str()
  // Auxiliary function to translate an ID code to a string, 
  // for constructing antenna names
    
  string id2str(int id) const {
    // Vector bosons
    if (id == 21) return "g";
    else if (id == 22) return "gamma";
    else if (id == 23) return "Z";
    else if (abs(id) == 24) return "W";
    // Fermions
    else if (id >= 1 && id <= 10) return "q";
    else if (-id >= 1 && -id <= 10) return "qbar";
    else if (id >= 11 && id <= 20 && id%2 == 1) return "l-"; 
    else if (id >= 11 && id <= 20 && id%2 == 0) return "nu"; 
    else if (-id >= 11 && -id <= 20 && id%2 == 1) return "l+"; 
    else if (-id >= 11 && -id <= 20 && id%2 == 0) return "nubar"; 
    // Octet fermion: use gluino
    else if (id == 1000021) return "~g";
    // Charged scalar: use H+
    else if (id == 37) return "H+";
    else if (id == -37) return "H-";
    // Colored scalars: use squarks
    else if (id >= 1000000 && id <= 1000010) return "~q";
    else if (-id >= 1000000 && -id <= 1000010) return "~q*";
    // Unknown particle
    else return "X";
  }
  
  //----------------------------------------------------------------------
  // List of PROTECTED data members
  // (Not private, so can be accessed by inheriting classes)

 protected:
    
  // Is initialized
  bool isInitPtr, isInit;
    
  // Charge factor, kinematics map, and subleading-colour treatment
  double chargeFacSav;
  int kineMapTypeSav, modeSLC;

  // The alpha collinear-partitioning parameter
  double alphaSav;
    
  // Shorthand for commonly used variable(s)
  double term;

  // Map to tell whether a given helicity value maps to L- and/or R-handed
  // Defined by constructor and not to be changed dynamically.
  map<int,bool> LH;
  map<int,bool> RH;

  // Verbose level
  int verbose;

  // Pointers to Vincia classes
  Functions* functionsPtr;

  // Pointers to Pythia-8 classes
  ParticleData* particleDataPtr;
  Rndm*         rndmPtr;
  Settings*     settingsPtr;  
    
};

//==========================================================================

// The AntennaISR class:
// Base class implementation for Initial-Final and Initial-Initial
// Antenna objects
// Note: unlike the FF ones, these antenna functions will be dimensionful

class AntennaISR {
    
 public:
    
  // Constructor
  AntennaISR() {
    // Init flags
    isInitPtr           = false;
    isInit              = false;
    // Default values (reset by init)
    verbose             = 1;
    chargeFacSav     = 0.0;
    alphaSav            = 0.0;
    preFacFiniteTermSav = 0.0;
    antMinSav           = 0.0;
    // Initialize LH and RH maps (DO NOT CHANGE)
    RH[9]  = true;
    RH[1]  = true;
    RH[-1] = false;
    LH[9]  = true;
    LH[1]  = false;
    LH[-1] = true;
  };
    
  // Destructor 
  virtual ~AntennaISR() {};
    
  //----------------------------------------------------------------------
  // List of VIRTUAL functions to be redefined by inheriting classes

  // Names of this antenna, for VINCIA, and for humans
  virtual string vinciaName()   const {return "Vincia:AntennaISR";};

  // Parton types AB -> ajb for II
  //              AK -> ajk for IF
  virtual int idA()    const { return 0; }
  virtual int idB()    const { return 0; }
  virtual int ida()    const { return 0; }
  virtual int idb()    const { return 0; }
  virtual int idj()    const { return 0; }

  // The antenna function
  virtual double antennaFunction(double s1j, double sj2, double S12, 
    int h1Old=9, int h2Old=9, 
    int h1=9, int h2=9, int h3=9);
    
  // Optional implementation of the AP kernels corresponding
  // to the two sides of the antenna, for collinear-limit checks.
  // NOTE AP kernels are implemented such that they are easily
  // comparable to the antennae, ie alpha/2pi*C vs alpha/4pi*C results
  // in a factor 1 for CF vs 2CF and TR/2 vs TR, but AP get multiplied
  // by 2 for CA vs CA. Colour factors: CA = 3, CF = 4/3, TR = 1.
  // Also additional factors of 1/2 are possible (taking into
  // account shared singularities)
  virtual double AltarelliParisi(double s1j, double sj2, double S12, 
    int h1Old=9, int h2Old=9, 
    int h1=9, int h2=9, int h3=9);

  //----------------------------------------------------------------------
  // List of VIRTUAL functions that MAY be redefined (if needed)

  // Default initialization
  virtual bool init();
      
  // Construct baseName from idA, idB, and idj
  virtual string baseName() const {
    string name = id2str(idj()) + "/" + id2str(idA()) + id2str(idB());
    return name;
  }

  // Functions to get AP energy fractions
  virtual double z1(double, double, double) { return -1.0; }
  virtual double z2(double, double, double) { return -1.0; }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return false; }

  // Function to check singularities, positivity, etc.
  virtual bool check();

  //----------------------------------------------------------------------
  // List of NON-VIRTUAL functions (i.e., that may not be redefined)

  // Initialize pointers:
  //    ParticleData : required for masses, color charges, etc
  //    Rndm : required to generate random test points to check antenna
  //    Settings : required for all coefficients, switches, etc.
  void initPtr(Functions* functionsPtrIn, ParticleData* particleDataPtrIn, 
    Rndm* rndmPtrIn, Settings* settingsPtrIn) {
    functionsPtr    = functionsPtrIn;
    particleDataPtr = particleDataPtrIn; 
    rndmPtr         = rndmPtrIn;
    settingsPtr     = settingsPtrIn; 
    isInitPtr       = true;
    isInit          = false;
  };

  // -------
  // Functions

  // Get parameters
  double getChargeFac() { return chargeFacSav; }
  double getAlpha() { return alphaSav; }
  
  //----------------------------------------------------------------------
  // id2str()
  // Auxiliary function to translate an ID code to a string, 
  // for constructing antenna names
    
  string id2str(int id) const {
    // Vector bosons
    if (id == 21) return "g";
    else if (id == 22) return "gamma";
    else if (id == 23) return "Z";
    else if (abs(id) == 24) return "W";
    // Fermions
    else if (id >= 1 && id <= 10) return "q";
    else if (-id >= 1 && -id <= 10) return "qbar";
    else if (id >= 11 && id <= 20 && id%2 == 1) return "l-"; 
    else if (id >= 11 && id <= 20 && id%2 == 0) return "nu"; 
    else if (-id >= 11 && -id <= 20 && id%2 == 1) return "l+"; 
    else if (-id >= 11 && -id <= 20 && id%2 == 0) return "nubar"; 
    // Octet fermion: use gluino
    else if (id == 1000021) return "~g";
    // Charged scalar: use H+
    else if (id == 37) return "H+";
    else if (id == -37) return "H-";
    // Colored scalars: use squarks
    else if (id >= 1000000 && id <= 1000010) return "~q";
    else if (-id >= 1000000 && -id <= 1000010) return "~q*";
    // Unknown particle
    else return "X";
  }
  
  //----------------------------------------------------------------------
  // List of PROTECTED data members
  // (Not private, so can be accessed by inheriting classes)

 protected:
    
  // Is initialized
  bool isInitPtr, isInit;  

  // Charge factor
  double chargeFacSav;
    
  // Prefactor of finite term and ant minimum
  double preFacFiniteTermSav;
  bool   isMinVar;
  double antMinSav;
    
  // The alpha collinear-partitioning variable
  double alphaSav;

  // Shorthand for commonly used variable(s)
  double term;

  // Map to tell whether a given helicity value maps to L- and/or R-handed
  // Defined by constructor and not to be changed dynamically.
  map<int,bool> LH;
  map<int,bool> RH;

  // Verbose level
  int verbose;

  // Pointers to Vincia classes
  Functions* functionsPtr;

  // Pointers to Pythia-8 classes
  ParticleData* particleDataPtr;
  Rndm*         rndmPtr;
  Settings*     settingsPtr;  
    
};

//==========================================================================

// Derived classes for Final-Final Antennae

//==========================================================================

// class QQEmitFF 

class QQEmitFF : public AntennaFSR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const {return "Vincia:QQEmitFF";};

  // Functions needed by soft- and collinear-limit checks (AB -> IJK)
  virtual int idA()    const {return 1;}
  virtual int idB()    const {return -1;}
  virtual int idj()    const {return 21;}

  // The (dimensionless) antenna function
  virtual double antennaFunction(double yij, double yjk, 
    double mui=0.0, double muj=0.0, double muk=0.0,
    int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);
    
  // Function to give Altarelli-Parisi limits of this antenna
  // Defined as PI/yij + PK/yjk, i.e., equivalent to antennae.
  virtual double AltarelliParisi(double yij, double yjk, double,double,double,
    int hA=9, int hB=9, int hI=9, int hJ=9, int hK=9){
    if (hA != hI || hB != hK) return -1.;
    else return functionsPtr->Pq2qg(zi(yij,yjk),hA,hI,hJ)/yij 
           + functionsPtr->Pq2qg(zk(yij,yjk),hB,hK,hJ)/yjk;
  }

  // Additional data members specific to this class
 private:

};

//==========================================================================

// class QGEmitFF

class QGEmitFF : public AntennaFSR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const {return "Vincia:QGEmitFF";};

  // Parton types (AB -> IJK): needed by soft- and collinear-limit checks 
  virtual int idA()    const {return 1;}
  virtual int idB()    const {return 21;}
  virtual int idj()    const {return 21;}

  // The (dimensionless) antenna function
  virtual double antennaFunction(double yij, double yjk, 
    double mui=0.0, double muj=0.0, double muk=0.0,
    int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);

  // Function to give Altarelli-Parisi limits of this antenna
  virtual double AltarelliParisi(double yij, double yjk,
    double, double, double, int hA=9, int hB=9, 
    int hI=9, int hJ=9, int hK=9) {
    double sum(0.);
    if (hI != hA) return -1;
    if (hK == hB) sum += functionsPtr->Pq2qg(zi(yij,yjk),hA,hI,hJ)/yij;
    sum += functionsPtr->Pg2gg(zk(yij,yjk),hB,hK,hJ)/yjk;
    return sum;
  }

  // Additional data members specific to this class
 private:
    
};

//==========================================================================

// class GGEmitFF 

class GGEmitFF : public AntennaFSR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const {return "Vincia:GGEmitFF";};

  // Parton types (AB -> IJK): needed by soft- and collinear-limit checks 
  virtual int idA()    const {return 21;}
  virtual int idB()    const {return 21;}
  virtual int idj()    const {return 21;}

  // The (dimensionless) antenna function
  virtual double antennaFunction(double yij, double yjk, 
    double mui=0.0, double muj=0.0, double muk=0.0,
    int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);

  // Function to give Altarelli-Parisi limits of this antenna
  virtual double AltarelliParisi(double yij, double yjk,
    double, double, double, int hA=9, int hB=9, 
    int hI=9, int hJ=9, int hK=9) {
    double sum(0.);
    if ( hB == hK ) sum += functionsPtr->Pg2gg(zi(yij,yjk),hA,hI,hJ)/yij;
    if ( hA == hI ) sum += functionsPtr->Pg2gg(zk(yij,yjk),hB,hK,hJ)/yjk;
    return sum;
  }

  // Additional data members specific to this class
 private:

};

//==========================================================================

// class QGSplitFF

class QGSplitFF : public AntennaFSR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const {return "Vincia:QGSplitFF";};

  // Parton types (AB -> IJK): needed by soft- and collinear-limit checks 
  virtual int idA()    const {return 1;}
  virtual int idB()    const {return 21;}
  virtual int idj()    const {return -2;}

  // The (dimensionless) antenna function
  virtual double antennaFunction(double yij, double yjk, 
    double mui=0.0, double muj=0.0, double muk=0.0,
    int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);

  // Function to give Altarelli-Parisi limits of this antenna
  virtual double AltarelliParisi(double yij, double yjk,
    double, double, double, int hA=9, int hB=9, 
    int hI=9, int hJ=9, int hK=9) {
    if (hA != hI) return 0.0; 
    else return functionsPtr->Pg2qq(zk(yij,yjk),hB,hK,hJ)/yjk;
  }

  // Additional data members specific to this class
 private:

};

//==========================================================================

// class GGSplitFF 
// Note: function coded is for g g -> g qbar q
// Vincia automatically includes the other permutation

class GGSplitFF : public AntennaFSR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const {return "Vincia:GGSplitFF";};

  // Parton types (AB -> IJK): needed by soft- and collinear-limit checks 
  virtual int idA()    const {return 21;}
  virtual int idB()    const {return 21;}
  virtual int idj()    const {return -2;}

  // The (dimensionless) antenna function
  virtual double antennaFunction(double yij, double yjk, 
    double mui=0.0, double muj=0.0, double muk=0.0,
    int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);

  // Function to give Altarelli-Parisi limits of this antenna
  virtual double AltarelliParisi(double yij, double yjk,
    double, double, double, int hA=9, int hB=9, 
    int hI=9, int hJ=9, int hK=9) {
    if (hA != hI) return 0.;
    else return functionsPtr->Pg2qq(zk(yij,yjk),hB,hK,hJ)/yjk;
  }

  // Additional data members specific to this class
 private:

};
  
//==========================================================================

// Derived classes for Initial-Initial Antennae

//==========================================================================

// class QQEmitII

class QQEmitII : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:QQemitII"; };

  // Parton types AB -> ajb with A,B,a,b initial and j final
  virtual int idA()    const { return 1; }
  virtual int idB()    const { return -1; }
  virtual int ida()    const { return 1; }
  virtual int idb()    const { return -1; }
  virtual int idj()    const { return 21; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjb, double sAB, 
    int hA=9, int hB=9, 
    int ha=9, int hj=9, int hb=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjb, double sAB, 
    int hA=9, int hB=9, 
    int ha=9, int hj=9, int hb=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double    , double sjb, double sAB) { return sAB/(sAB+sjb); }
  virtual double z2(double saj, double    , double sAB) { return sAB/(sAB+saj); }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return true; }

  // Additional data members specific to this class
 private:

};

//--------------------------------------------------------------------------
// class GQEmitII

class GQEmitII : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GQemitII"; };

  // Parton types AB -> ajb with A,B,a,b initial and j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 1; }
  virtual int ida()    const { return 21; }
  virtual int idb()    const { return 1; }
  virtual int idj()    const { return 21; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjb, double sAB, 
    int hA=9, int hB=9, 
    int ha=9, int hj=9, int hb=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjb, double sAB, 
    int hA=9, int hB=9, 
    int ha=9, int hj=9, int hb=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double    , double sjb, double sAB) { return sAB/(sAB+sjb); }
  virtual double z2(double saj, double    , double sAB) { return sAB/(sAB+saj); }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return true; }

  // Additional data members specific to this class
 private:
    
};

//--------------------------------------------------------------------------
// class GGEmitII 

class GGEmitII : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GGemitII"; };

  // Parton types AB -> ajb with A,B,a,b initial and j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 21; }
  virtual int ida()    const { return 21; }
  virtual int idb()    const { return 21; }
  virtual int idj()    const { return 21; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjb, double sAB, 
    int hA=9, int hB=9, 
    int ha=9, int hj=9, int hb=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjb, double sAB, 
    int hA=9, int hB=9, 
    int ha=9, int hj=9, int hb=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double    , double sjb, double sAB) { return sAB/(sAB+sjb); }
  virtual double z2(double saj, double    , double sAB) { return sAB/(sAB+saj); }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return true; }

  // Additional data members specific to this class
 private:

};

//--------------------------------------------------------------------------
// class QXSplitII
// NOTE Splitting in the forwards sense aka quark backwards evolving
//       to a gluon and emitting a antiquark in the final state

class QXSplitII : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:QXsplitII"; };

  // Parton types AB -> ajb with A,B,a,b initial and j final
  virtual int idA()    const { return 1; }
  virtual int idB()    const { return 0; }
  virtual int ida()    const { return 21; }
  virtual int idb()    const { return 0; }
  virtual int idj()    const { return -1; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjb, double sAB, 
    int hA=9, int hB=9, 
    int ha=9, int hj=9, int hb=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjb, double sAB, 
    int hA=9, int hB=9, 
    int ha=9, int hj=9, int hb=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double    , double sjb, double sAB) { return sAB/(sAB+sjb); }
  virtual double z2(double    , double    , double    ) { return -1.0; }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return true; }

  // Additional data members specific to this class
 private:

};

//--------------------------------------------------------------------------
// class GXConvII
// NOTE Gluon evolves backwards into a quark and emits a quark
//       in the final state

class GXConvII : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GXconvII"; };

  // Parton types AB -> ajb with A,B,a,b initial and j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 0; }
  virtual int ida()    const { return 2; }
  virtual int idb()    const { return 0; }
  virtual int idj()    const { return 2; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjb, double sAB, 
    int hA=9, int hB=9, 
    int ha=9, int hj=9, int hb=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjb, double sAB, 
    int hA=9, int hB=9, 
    int ha=9, int hj=9, int hb=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double    , double sjb, double sAB) { return sAB/(sAB+sjb); }
  virtual double z2(double    , double    , double    ) { return -1.0; }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return true; }

  // Additional data members specific to this class
 private:

};

//==========================================================================

// Derived classes for Initial-Final Antennae

//==========================================================================

// class QQEmitIF

class QQEmitIF : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:QQemitIF"; };

  // Parton types AB -> ajb with A,a initial and B,b,j final
  virtual int idA()    const { return 1; }
  virtual int idB()    const { return -1; }
  virtual int ida()    const { return 1; }
  virtual int idb()    const { return -1; }
  virtual int idj()    const { return 21; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double    , double sjk, double sAK) { return sAK/(sAK+sjk); }
  virtual double z2(double saj, double    , double sAK) { return (sAK-saj)/sAK; }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return false; }

  // Additional data members specific to this class
 private:

};

//--------------------------------------------------------------------------
// class QGEmitIF

class QGEmitIF : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:QGemitIF"; };

  // Parton types AB -> ajb with A,a initial and B,b,j final
  virtual int idA()    const { return 1; }
  virtual int idB()    const { return 21; }
  virtual int ida()    const { return 1; }
  virtual int idb()    const { return 21; }
  virtual int idj()    const { return 21; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double    , double sjk, double sAK) { return sAK/(sAK+sjk); }
  virtual double z2(double saj, double    , double sAK) { return (sAK-saj)/sAK; }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return false; }

  // Additional data members specific to this class
 private:
    
};

//--------------------------------------------------------------------------
// class GQEmitIF

class GQEmitIF : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GQemitIF"; };

  // Parton types AB -> ajb with A,a initial and B,b,j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 1; }
  virtual int ida()    const { return 21; }
  virtual int idb()    const { return 1; }
  virtual int idj()    const { return 21; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double    , double sjk, double sAK) { return sAK/(sAK+sjk); }
  virtual double z2(double saj, double    , double sAK) { return (sAK-saj)/sAK; }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return false; }

  // Additional data members specific to this class
 private:
    
};

//--------------------------------------------------------------------------
// class GGEmitIF 

class GGEmitIF : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GGemitIF"; };

  // Parton types AB -> ajb with A,a initial and B,b,j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 21; }
  virtual int ida()    const { return 21; }
  virtual int idb()    const { return 21; }
  virtual int idj()    const { return 21; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double    , double sjk, double sAK) { return sAK/(sAK+sjk); }
  virtual double z2(double saj, double    , double sAK) { return (sAK-saj)/sAK; }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return false; }

  // Additional data members specific to this class
 private:

};

//--------------------------------------------------------------------------
// class QXSplitIF
// NOTE Splitting in the forwards sense aka quark backwards evolving
//       to a gluon and emitting a antiquark in the final state

class QXSplitIF : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:QXsplitIF"; };

  // Parton types AB -> ajb with A,a initial and B,b,j final
  virtual int idA()    const { return 1; }
  virtual int idB()    const { return 0; }
  virtual int ida()    const { return 21; }
  virtual int idb()    const { return 0; }
  virtual int idj()    const { return -1; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double, double sjk, double sAK) { return sAK/(sAK+sjk); }
  virtual double z2(double, double    , double    ) { return -1.0; }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return false; }

  // Additional data members specific to this class
 private:

};

//--------------------------------------------------------------------------
// class GXConvIF
// NOTE Gluon evolves backwards into a quark and emits a quark
//       in the final state

class GXConvIF : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GXconvIF"; };

  // Parton types AB -> ajb with A,a initial and B,b,j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 0; }
  virtual int ida()    const { return 2; }
  virtual int idb()    const { return 0; }
  virtual int idj()    const { return 2; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double, double sjk, double sAK) { return sAK/(sAK+sjk); }
  virtual double z2(double, double    , double    ) { return -1.0; }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return false; }

  // Additional data members specific to this class
 private:

};

//--------------------------------------------------------------------------
// class XGSplitIF
// NOTE Gluon splitting in the final state

class XGSplitIF : public AntennaISR {

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:XGsplitIF"; };

  // Parton types AB -> ajb with A,a initial and B,b,j final
  virtual int idA()    const { return 0; }
  virtual int idB()    const { return 21; }
  virtual int ida()    const { return 0; }
  virtual int idb()    const { return 1; }
  virtual int idj()    const { return -1; }

  // The antenna function
  virtual double antennaFunction(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);
    
  // AP splitting kernel
  virtual double AltarelliParisi(double saj, double sjk, double sAK, 
    int hA=9, int hK=9, 
    int ha=9, int hj=9, int hk=9);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double z1(double    , double, double    ) { return -1.0; }
  virtual double z2(double saj, double, double sAK) { return (sAK-saj)/sAK; }
    
  // Function to tell if this is an II antenna
  virtual bool isIIant() { return false; }

  // Additional data members specific to this class
 private:

};
  
//==========================================================================

// The AntennaSetFSR class:
// Contains a list of FSR Antenna Functions

class AntennaSetFSR {
  
 public:
  
  // Destructor: delete the antennae
  ~AntennaSetFSR(){
    for (vector<AntennaFSR*>::iterator antPtrIt = antPtrs.begin(); 
         antPtrIt != antPtrs.end(); ++antPtrIt) delete *antPtrIt;
  }
  // Initialize pointers
  void initPtr(Functions* functionsPtrIn, ParticleData* particleDataPtrIn, 
    Rndm* rndmPtrIn, Settings* settingsPtrIn) {
    functionsPtr    = functionsPtrIn;
    particleDataPtr = particleDataPtrIn; 
    rndmPtr         = rndmPtrIn;
    settingsPtr     = settingsPtrIn; 
    isInitPtr       = true;
  }
  
  // Initialize antenna set (optionally with min or max variation)
  void init();
  
  // Function to chek if an antenna with the given index exists
  bool exists(int iAntIn) {
    return ( iAntIn<0 || iAntIn >= int(antPtrs.size()) ) ? false : true;
  }
  
  // Gets an antenna iAntIn from the AntennaSet
  AntennaFSR* getAnt(vector<AntennaFSR*>::size_type iAntIn) { 
    if (iAntIn <= antPtrs.size()) {
      return antPtrs[iAntIn];
    }
    else return NULL;
  }

  // Get Vincia name (e.g., "Vincia:QQ:emit")
  string vinciaName(vector<AntennaFSR*>::size_type iAntIn) {
    if (iAntIn <= antPtrs.size()) {
      return antPtrs[iAntIn]->vinciaName();
    }
    else return "noVinciaName";        
  }

  // Get human name (e.g., "g/qq")
  string humanName(vector<AntennaFSR*>::size_type iAntIn) {
    if (iAntIn <= antPtrs.size()) {
      return antPtrs[iAntIn]->humanName();
    }
    else return "noHumanName";        
  }

 private:

  // use a vector of AntennaFSR pointers, create them with new on 
  // initialization
  vector<AntennaFSR*> antPtrs;

  // Verbose level
  int verbose;

  // Pointers to Vincia classes
  Functions*    functionsPtr;

  // Pointers to Pythia-8 classes, needed to initialise antennae
  bool isInitPtr;
  ParticleData* particleDataPtr;
  Rndm*         rndmPtr;
  Settings*     settingsPtr;  

}; // End class AntennaSetFSR
  
//==========================================================================

// The AntennaSetISR class:
// Added by NF to keep antenna sets for ISR shower separate from
// from the FSR ones.
  
class AntennaSetISR {
    
 public:
    
  // Destructor: delete the antennae
  ~AntennaSetISR(){
    for (vector<AntennaISR*>::iterator antPtrIt = antPtrs.begin(); 
         antPtrIt != antPtrs.end(); ++antPtrIt) delete *antPtrIt;
  }
  // Initialize pointers
  void initPtr(Functions* functionsPtrIn, ParticleData* particleDataPtrIn, 
    Rndm* rndmPtrIn, Settings* settingsPtrIn) {
    functionsPtr    = functionsPtrIn;
    particleDataPtr = particleDataPtrIn; 
    rndmPtr         = rndmPtrIn;
    settingsPtr     = settingsPtrIn; 
    isInitPtr       = true;
  }
 
  // Initialize antenna set
  void init();

  // Function to chek if an antenna with the given index exists
  bool exists(int iAntIn) {
    return ( iAntIn<0 || iAntIn >= int(antPtrs.size()) ) ? false : true;
  }
    
  // Gets an antenna iAntIn from the AntennaSetISR
  AntennaISR* getAnt(vector<AntennaISR*>::size_type iAntIn) { 
    if (iAntIn <= antPtrs.size()) return antPtrs[iAntIn];
    else return NULL;
  }

  // Get Vincia name (e.g., "Vincia:QQemitII")
  string vinciaName(vector<AntennaISR*>::size_type iAntIn) {
    if (iAntIn <= antPtrs.size()) return antPtrs[iAntIn]->vinciaName();
    else return "noVinciaName";        
  }

 private:

  // use a vector of AntennaISR pointers, create them with new on 
  // initialization
  vector<AntennaISR*> antPtrs;

  // Verbose level
  int verbose;

  // Pointers to Vincia classes
  Functions*    functionsPtr;

  // Pointers to Pythia-8 classes, needed to initialise antennae
  bool isInitPtr;
  ParticleData* particleDataPtr;
  Rndm*         rndmPtr;
  Settings*     settingsPtr;  

}; // End class AntennaSetISR

} // End Vincia namespace

#endif
