// SpaceAntShower.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2017 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the SpaceAntennaShower class
  
#ifndef VINCIASPACEANTSHOWER_H
#define VINCIASPACEANTSHOWER_H

#include "Pythia8/SpaceShower.h"
#include "Vincia/VinciaCommon.h"
#include "Vincia/VinciaWeights.h"

#include "Vincia/AntennaFunctions.h"
#include "Vincia/Branchelementals.h"
#include "Vincia/Reconnections.h"
#include "Vincia/Resolution.h"
#include "Vincia/TrialGeneratorsISR.h"
#include "Vincia/MatrixElementCorrections.h"
#include "Vincia/MergingHooks.h"
#include "Vincia/Tools.h"

namespace Vincia {
 
// Forward declarations
class TimeAntShower;
class MatrixElementCorrections;
class VinciaMergingHooks;
class VinciaMerging;

// Classes contained in this file 
class SpaceAntShower; 

//**************************************************************************

// The SpaceAntShower class
// Main shower class for initial-state (II and IF) antenna showers
// Inherits from SpaceShower in Pythia 8 so can be used as alternative to 
// SpaceShower. Added by NF.

class SpaceAntShower : public Pythia8::SpaceShower {

  // Allow VinciaPlugin and TimeAntShower to access private information
  friend class VinciaPlugin;
  friend class TimeAntShower;

 public:

  // ----------------------------------------------------------------------

  // CONSTRUCTOR AND DESTRUCTOR 

  // Constructor.
  SpaceAntShower() : isInit(false), headerIsPrinted(false) { };

  // Destructor.
  virtual ~SpaceAntShower() { };

  // -----------------------------------------------------------------------
    
  // FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING 

  // Initialize generation. Possibility to force re-initialization by hand.
  virtual void init( BeamParticle* beamAPtrIn, BeamParticle* beamBPtrIn);

  // Possible limitation of first emission.
  virtual bool limitPTmax( Event& event, double Q2Fac = 0., double Q2Ren = 0.);

  // Potential enhancement factor of pTmax scale for hardest emission.
  // Used if limitPTmax = true.
  virtual double enhancePTmax() const {
    return 1.0;
  }

  // Prepare system for evolution; identify ME.
  virtual void prepare( int iSys, Event& event, bool limitPTmaxIn = false);

  // Update dipole list after each ISR emission.
  virtual void update( int iSys, Event& event, bool hasWeakRad = false);

  // Select next pT in downwards evolution.
  virtual double pTnext( Event& event, double pTbegAll, double pTendAll,
    int nRadIn = -1, bool doTrialIn = false);

  // ME corrections and kinematics that may give failure.
  virtual bool branch( Event& event);
    
  // Utility to print dipole list; for debug mainly.
  virtual void list() const;
    
  // Functions inherited for merging.

  // Return clustered event.
  vector<Event> clusteredAll( const Event&, int, int, int, string);
  Event clustered( const Event& event, int iRa, int iEm, int iRe, string
    name ) {
    vector<Event> clusteredEvents = clusteredAll(event, iRa, iEm, iRe, name);
    return ( clusteredEvents.size() > 0 ? clusteredEvents[0] : Event() );
  }

  // Return the evolution variable.
  vector< map<string, double> > getStateVariablesAll( const Event&, int, int,
    int, string);
  map<string, double> getStateVariables( const Event& event, int iRa, int iEm, 
    int iRe, string name) { 
    vector< map<string, double> > scales = getStateVariablesAll(event, iRa, iEm,
      iRe, name);
    return ( scales.size() > 0 ? scales[0] : map<string, double>() );
  }

  // Check if attempted clustering is handled by spacelike shower.
  bool isSpacelike( const Event&, int, int, int, string);

  // Return a string identifier of a splitting.
  vector<string> getSplittingName( const Event&, int, int, int);

  // Return the splitting probability.
  vector<double> getSplittingProbAll( const Event&, int, int, int, string);
  double getSplittingProb( const Event& event, int iRa, int iEm, int iRe, 
    string name) { 
    vector<double> splittingProbs = getSplittingProbAll(event, iRa, iEm, iRe, 
      name);
    return ( splittingProbs.size() > 0 ? splittingProbs[0] : 0.0 );
  }

  // TODO
  //bool allowedSplitting( const Event&, int, int);

  // TODO
  //vector<int> getRecoilers( const Event&, int, int, string);

  // END OF FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING

  // -----------------------------------------------------------------------

  // PUBLIC NON-INHERITED MEMBERS

  // Return matrix-element correction.
  double getMEC( const Event& event, int nStep=-1);

  // Return scales, alphaS values, etc in case of only unordered shower paths.
  map<string, double> getUnorderedInfos( const Event& event);

  // Return the Born probability (= ME).
  double getBornProb( const Event& event);

  // Return the Born scale for merging.
  double getBornScale( const Event& event);

  // Return the shower starting scale for pure shower.
  double getShowerStartingScale( const Event& event); 

  // Update antennae if FSR changed system in prepare
  void updateAfterFSR(int iSys, Event& event, map<int,int> iChanged);

  // Wrapper function to return a specific antenna inside antenna set 
  AntennaISR* getAnt(int iAnt) { return antSetPtr->getAnt(iAnt); }

  // Initialize Pointers to Antenna sets
  void initAntPtr(AntennaSetISR* antSetIn) { antSetPtr = antSetIn; }

  // Print out the antennae
  void printAnts(int iSys);

  // Print header information (version, settings, parameters, etc)
  void header();

  // Print final statistics information
  void printInfo(bool pluginCall = false );

  // Normalization factor
  double getEvNormFac(int iAntPhys) {
    double evNorm = (iAntPhys<iQQemitIF ? evNormIIpT2 : evNormIFpT2);
    if (!evAllPT) {
      if (iAntPhys == iQXsplitII || iAntPhys == iQXsplitIF) evNorm = evNormMsplitI;
      else if (iAntPhys == iGXconvII || iAntPhys == iGXconvIF) evNorm = evNormMconvI;
      else if (iAntPhys == iXGsplitIF) evNorm = evNormMsplitF;
    }
    return evNorm;
  }

  // Evolution Windows: Phase space region boundaries
  int getRegion(double q) {
    for (int i=1; i<(int)regMinScalesSav.size(); i++)
      if (q < regMinScalesSav[i]) return i-1;
    return (int)regMinScalesSav.size()-1;
  }

  // Evolution Windows: Phase space region boundaries
  double getQmin(int iRegion) {
    iRegion = max(0,iRegion);
    iRegion = min(iRegion,(int)regMinScalesSav.size()-1);
    return regMinScalesSav[iRegion];
  }

  // Number of active flavors 
  int getNf(int iRegion) {
    if (iRegion <= 1) return 3;
    else if (iRegion <= 2) return 4;
    else if (iRegion <= 4) return 5;
    else return 6;
  }

  // Pointer to global AlphaS instance
  AlphaStrong* alphaSptr;

  // Lambda value
  double getLambda(int nFin) {
    if (nFin <= 3) return alphaSptr->Lambda3();
    else if (nFin <= 4) return alphaSptr->Lambda4();
    else if (nFin <= 5) return alphaSptr->Lambda5();
    else return alphaSptr->Lambda6();
  }

  // Functions to initialize, set and get headroomFacs
  void initHeadroom(double valEmit, double valSplitII, double valSplitIF, 
    double valConvII, double valConvIF, double valSplitF) {
    for (int iTrialGen=0; iTrialGen<=13; iTrialGen++) {
      double valNow = 1.0;
      if ( iTrialGen<3 || (iTrialGen>4 && iTrialGen<11) ) valNow = valEmit;
      else if (iTrialGen == 3)  valNow = valSplitII;
      else if (iTrialGen == 11) valNow = valSplitIF;
      else if (iTrialGen == 4)  valNow = valConvII;
      else if (iTrialGen == 13) valNow = valConvIF;
      else if (iTrialGen == 12) valNow = valSplitF;
      // Loop over regions
      for (int iReg=0; iReg<=20; iReg++) {
        // Loop over number of gluons
        for (int nGin=0; nGin<=8; nGin++) {
          // Loop over number of quarks
          for (int nQin=0; nQin<=8; nQin++) {
            headroom[iTrialGen][iReg][nGin][nQin]             = valNow;
            headroomMECs[iTrialGen][iReg][nGin][nQin]         = 2.0*valNow;
            headroomOptimize[iTrialGen][iReg][nGin][nQin]     = optimizeHeadroom;
            headroomMECsOptimize[iTrialGen][iReg][nGin][nQin] = optimizeHeadroom;
          }
        }
      }
    }
  }
  double getHeadroom(int iTrialGen, int iReg, int nGin, int nQin,
    int iSys, double qMin) {
    if (!iRegDependence) iReg = 1;
    else if (iReg == -1) iReg = getRegion(qMin);
    if (iTrialGen < 0  || iReg < 0  || nGin < 0 || nGin > 8 || 
      iTrialGen > 13 || iReg > 20 || nQin < 0 || nQin > 8) return 1.0;
    if ((nBranch[iSys]+1<=matchingLO) && hardSys[iSys])
      return headroomMECs[iTrialGen][iReg][nGin][nQin];
    else return headroom[iTrialGen][iReg][nGin][nQin];
  }
  void setHeadroom(int iTrialGen, int iReg, int nGin, int nQin,
    int iSys, double qMin, double scale) {
    if (!iRegDependence) iReg = 1;
    else if (iReg == -1) iReg = getRegion(qMin);
    if (iTrialGen < 0  || iReg < 0  || nGin < 0 || nGin > 8 || 
      iTrialGen > 13 || iReg > 20 || nQin < 0 || nQin > 8) return;
    bool MEC      = ((nBranch[iSys]+1<=matchingLO) && hardSys[iSys]);
    double maxVal = 5.0;
    double oldVal = (MEC ? headroomMECs[iTrialGen][iReg][nGin][nQin]
      : headroom[iTrialGen][iReg][nGin][nQin]);
    // Switch off further optimization if increasing headroom factor
    if ( scale > 1.0 ) {
      if (MEC) {
        headroomMECsOptimize[iTrialGen][iReg][nGin][nQin] = false;
        headroomMECs[iTrialGen][iReg][nGin][nQin] = min(scale*oldVal,maxVal);
      } else {
        headroomOptimize[iTrialGen][iReg][nGin][nQin] = false;
        headroom[iTrialGen][iReg][nGin][nQin] = min(scale*oldVal,maxVal);
      }
      if (verbose >= 5) {
        printOut("SpaceAntShower::setHeadroom","Increasing headroom at nG = "
          +num2str(nGin,2)+" nQ = "+num2str(nQin,2)+"; current "
          "headroom = "+num2str(oldVal));
      }
    } else {
      // Allow reduction of headroom if we're still optimizing
      if (MEC) {
        if (headroomMECsOptimize[iTrialGen][iReg][nGin][nQin])
          headroomMECs[iTrialGen][iReg][nGin][nQin] = scale*oldVal;
      } else {
        if (headroomOptimize[iTrialGen][iReg][nGin][nQin])
          headroom[iTrialGen][iReg][nGin][nQin] = scale*oldVal;
      }
    }
  }

  // Add trial functions to a Branchelemental
  void resetTrialGenerators(BranchelementalISR* trial);

  // Method to check if a gluon splitting in the initial state (to get rid of
  // heavy quarks) is still possible after the current branching
  bool checkHeavyQuarkPhaseSpace(vector<Particle> parts, int iSyst);
    
  // Method to check if we have heavy quark left after passing the evolution window
  bool heavyQuarkLeft(double qTrial);
    
  // Function to ask if a system is hard
  bool isSysHard(int iSys) {
    if (!isInit) return false;
    if ((int)hardSys.size() <= iSys) return false;
    return hardSys[iSys];
  }
    
  vector<double> getMasses() {
    vector<double> masses;
    masses.resize(0);
    masses.push_back(mt);
    masses.push_back(mtb);
    masses.push_back(mb);
    masses.push_back(mc);
    masses.push_back(ms);
    return masses;
  }

  // Information.
  // ---------------------------------------------------------------------------
  // Get number of systems
  int getNsys() {return nBranchISR.size();}
  // Get number of branchings in a system (return -1 if no such system)
  int getNbranch(int iSys = -1) {
    // iSys < 0 : return sum over all
    int n = 0;
    if (iSys < 0) for (int i=0; i<(int)nBranchISR.size(); ++i)
                    n += nBranchISR[iSys];
    else if (iSys < (int)nBranchISR.size()) 
      n = nBranchISR[iSys];
    else n = -1;
    return n;
  }
  // Get scale of branchings; use (0,1) for first branching in 1st system.
  // (Could be extended so that (i,0) would return starting scale for system i.)
  double getQbranch(int iSys, int iBranch) {
    if (iSys < 4 && iSys >= 0 && iBranch <= 10 && iBranch >= 1)
      return qBranch[iSys][iBranch];
    else return 0.;
  }
  double getPTphys(int iSys, int iBranch) {
    if (iSys < 4 && iSys >= 0 && iBranch <= 10 && iBranch >= 1)
      return pTphys[iSys][iBranch];
    else return 0.;
  }
      
  // Set verbose level
  void setVerbose(int verboseIn) { verbose = verboseIn; }

    
 private : 
    
  // -----------------------------------------------------------------------
    
  // PRIVATE MEMBERS
    
  // Generate trial branching kinematics and check physical phase space
  bool generateKinematics(BranchelementalISR* trialPtr, vector<Vec4>& pRec) {
    if (trialPtr->isII()) 
      return generateKinematicsII(trialPtr, pRec);
    else 
      return generateKinematicsIF(trialPtr);
  }
  bool generateKinematicsII(BranchelementalISR* trialPtr, vector<Vec4>& pRec);
  bool generateKinematicsIF(BranchelementalISR* trialPtr);
    
  // Main trial accept function
  bool acceptTrial(Event& event, BranchelementalISR* winnerPtr, 
    int iSysBranch, vector<Particle> partsNow, vector<int> antPos);

  // Method to assign colour flow
  bool assignColourFlow(Event& event, BranchelementalISR* trialPtr);

  // Initialized?
  bool isInit;
    
  // Beams
  int beamFrameType;
  BeamParticle* beamAPtr;
  BeamParticle* beamBPtr;
  double eBeamA, eBeamB, sbbSav;
    
  // Possibility to allow user veto of emission step.
  bool hasUserHooks, canVetoEmission;

  // Shower parameters
  int nGluonToQuarkI, nGluonToQuarkF;
  double cutoffScaleII, cutoffScaleIF;
  double evNormIIpT2, evNormIFpT2;
  double evNormMsplitI, evNormMconvI, evNormMsplitF;
  bool evAllPT;
  int nFlavZeroMass, massCorrectionLevel;
  bool treatAsMassless;

  // AlphaS parameters
  bool useCMW;
  int alphaSorder;
  double alphaSvalue, alphaSmax, alphaSmuFreeze, alphaSmuMin;
  double aSkMu2, aSkMu2SplitI, aSkMu2SplitF, aSkMu2Conv;
  double mt, mtb, ms, mb, mc;
  // Calculated values
  double mu2freeze, mu2min;
    
  // Trial Generators
  TrialIISoft    trialIISoft;
  TrialIIGCollA  trialIIGCollA;
  TrialIIGCollB  trialIIGCollB;
  TrialIISplitA  trialIISplitA;
  TrialIISplitB  trialIISplitB;
  TrialIIConvA   trialIIConvA;
  TrialIIConvB   trialIIConvB;
  TrialIFSoft    trialIFSoft;
  TrialIFQCollA  trialIFQCollA;
  TrialIFQCollK  trialIFQCollK;
  TrialIFGCollA1 trialIFGCollA1;
  TrialIFGCollA2 trialIFGCollA2;
  TrialIFGCollK  trialIFGCollK;
  TrialIFSplitA  trialIFSplitA;
  TrialIFSplitK  trialIFSplitK;
  TrialIFConvA   trialIFConvA;

  // Uncertainty Bands and Weights
  bool uncertaintyBands, allowReweighting;

  // Biasing switches and parameters
  bool biasInHard, biasInResDec, biasInMPI;
  double biasAll, biasBottom, biasCharm, biasCutoff;

  // Colour Reconnection parameters
  int  CRmode;
  bool CRjunctions;

  // Polarization global flag
  bool helicityShower;
  // Polarization and mass local flags
  bool isMassless, usePolarizationNow;
    
  // Desired matching multiplicities, level, regulator, regulator scales
  MatrixElementCorrections* matrixElementCorrectionsPtr;
  bool meCorrections, meCorrectionsNow;
  int  matchingLO, matchingNLO;
  int  matchingRegOrder, matchingRegShape;
  double matchingScale, matchingScaleRatio, matchingIRcutoff;
  bool matchingScaleIsAbs, matchingFullColor;
  bool Hinterference;

  // Pointer to Evolution Variable
  Resolution* resolutionPtr;

  // Pointer to VINCIA common, functions, clustering
  VinciaCommon* vinComPtr;
  Functions* functionsPtr;
  VinClu* vinCluPtr;

  // Pointer to weights class
  VinciaWeights* weightsPtr;

  // Pointer to VINCIA FSR shower
  TimeAntShower* FSRshowerPtr;

  // Pointer to VINCIA Reconnections object
  Reconnections* reconnectionsPtr;

  // New headroomFacs
  // first arg = iTrialGen, second arg = iRegion, third arg = nG, fourth arg = nG
  double headroom[14][21][9][9];
  bool   headroomOptimize[14][21][9][9];
  double headroomMECs[14][21][9][9];
  bool   headroomMECsOptimize[14][21][9][9];
  bool   optimizeHeadroom, iRegDependence;

  // Vector of antenna sets
  AntennaSetISR* antSetPtr;   

  // Total and MEC accept probability
  vector<double> Paccept;
  double PME;

  // Factorization scale and shower starting settings
  int    qMaxMatch;
  double q2fac, qFac;
  double q2maxFudge, q2maxFudgeMPI;

  // Parameters for merging and pointer to Vincia's merging
  bool doMergeSav;
  int  incomingType;
  VinciaMergingHooks* vinMergingHooksPtr;

  // Evolution windows
  vector<double> regMinScalesMtSav;
  vector<double> regMinScalesSav;
  vector<double> regMinScalesNow;

  // Current winner
  BranchelementalISR* winnerPtr;
  int indxWin;
  int iSysWin;
    
  // ---------------------------- Every system gets own copy -------------------------------- //
    
  // Is system a hard system?
  vector<bool> hardSys;
    
  // Some properties for MECs
  vector<int> MEclass, nQLneeded, Hint, additionalJets;

  // Saved particle state and number in event record
  vector< vector< Particle > > partsSav;
  vector< vector< int      > > indexSav;

  // Polarization local flags
  vector<bool> polarizedSys;

  // vector of dipoles (with trial branchings, 4 at most)
  vector< vector<BranchelementalISR> > branchelementals;

  // Save initial ISR starting scale system by system
  vector<double> Q2hat;

  // Count the number of branchings in the system
  vector<int> nBranch, nBranchISR;
    
  // Saved incoming guys
  vector<Particle> initialA;
  vector<Particle> initialB;
  double eAused;
  double eBused;
    
  // Count numbers of quarks and gluons
  vector<int> nG, nQQ;
    
  // ---------------------------- //

  // Statistics
  int nTrialsSum, nTrialsVeto;
  int nTrialsHull, nTrialsHullMassive;
  int nTrialsCutM, nTrialsCutQ;
  int nBornVeto, nBornPlusJetVeto;
  int nClosePSforHQ, nForceSplit;
  vector<int> nMCviols, nNegPas;
  vector<double> maxMCviols, minNegPas;
    
  // Rescue mechanism
  bool doRescue;
  int nRescue;
  double rescueMin;

  // Verbose setting, P>1 warning settings, and version number
  int verbose;
  double pAcceptFuss, pAcceptPT0;
  int nReportMCviol, nReportedMCviol;
  double version;
  bool headerIsPrinted;
  double qBranch[4][11], pTphys[4][11];
    
  // Counter for numbers of events
  long nAccepted, nSelected;
  int nVetoUserHooks, nFailHadLevel, nCallPythiaNext;
    
};
  
}

#endif
