// VinciaCommon.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2017 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the global auxiliary helper 
// functions for VINCIA, including the VinciaCommon class.

#ifndef VINCIACOMMON_H
#define VINCIACOMMON_H

// Include Pythia 8 headers and name space

#include "Pythia8/Event.h"           // Event
#include "Pythia8/StandardModel.h"   // AlphaStrong

#include <limits>

// ------------ GLOBAL VINCIA CONSTANTS AND FUNCTIONS --------------------
namespace Vincia {
  
  // Use the PYTHIA8 and STD namespaces while inside the VINCIA namespace.
  using namespace Pythia8;

  // Convenient typedef for unsigned integers.
  typedef unsigned int uint;

  // Global numerical precision target
  // (should be kept within a reasonable margin of machine precision)
  // (large values may result in strange results due to non-zero terms
  //  being artificially forced to zero)
  const double TINY     = 1.0e-9;
  const double TINYMASS = 1.0e-6;
  
  //**************************************************************************

  // Utilities for printing out VINCIA Messages
  void printErr(string,string);
  void printOut(string,string);
   
  // Utilities for converting data to strings
  string num2str(int,int width=4) ;
  string num2str(double,int width=9) ;
  string bool2str(bool, int width=2) ;
  
  // Utilities for checking if a file exists
  inline bool fileExists(const std::string& name) {
    if (FILE *file = fopen(name.c_str(), "r")) {
      fclose(file);
      return true;
    } else {
      return false;
    }   
  }
  
//***********************************************************************

  // VinciaCommon: 
  // Contains functions and variables shared by VinciaShower and VinciaMatching

  class VinciaCommon {
    
  public:
    
    // Constructor.
    VinciaCommon() {isInitPtr=false; isInit=false;}

    // Destructor.
    virtual ~VinciaCommon() {}

    // Initialize pointers
    bool initPtr(ParticleData* particleDataPtrIn, Settings* settingsPtrIn) {
      particleDataPtr = particleDataPtrIn; 
      settingsPtr     = settingsPtrIn; 
      isInitPtr       = true;
      return true;
    }

    // Initialize data members
    bool init();
  
    // Method to sort list of partons in Vincia colour order
    // Returns vector<int> with indices in following order:
    //   colourless incoming particles
    //   triplet-ngluon-antitriplet contractions
    //   ngluon rings
    //   colourless outgoing particles
    // No specific order is imposed on the relative ordering inside each
    // of the four classes.
    vector<int> colourSort(vector<Particle*>);

    // Function to check for the hadronization cutoff for a 
    // colour connected parton pair
    double mHadMin(const int, const int);
    
    // Function to check the event after each branching.
    // Added by NF to see if certain Pythia warnings/error are caused
    // by the shower.
    void showerChecks(Event& event, bool ISR);
    
    // Function to reset counters (print once every event for now)
    void resetCounters() {
      nUnkownPDG    = 0;
      nIncorrectCol = 0;
      nNAN          = 0;
      nVertex       = 0;
      nChargeCons   = 0;
      nMotDau       = 0;
      for (int i=0; i<2; i++) {
        nUnmatchedMass[i] = 0;
        nEPcons[i]        = 0;
      }
    }
    
    // Functions for Newton Solver
    double f(double t);
    double findRootSecant1D(double xmin, double xmax, double constant, int N);
    double findRoot1D(double xmin, double xmax, double constant, int N,
      double tol);
    double solveForT(double tMin, double tMax, double c);

    // Get number of active flavors at a given Q scale
    int getNf(double q) {
      if (q <= mc) return 3;
      else if (q <= mb) return 4;
      else if (q <= mt) return 5;
      else return 6;
    }

    // Get the shower starting scale
    double getShowerStartingScale( const Event& event, double sbbSav);

    // Verbosity level
    void setVerbose(int verboseIn) { verbose = verboseIn;}

    // Public data members: strong coupling in MSbar and CMW schemes,
    // user and default choices
    AlphaStrong alphaStrong;
    AlphaStrong alphaStrongCMW;
    AlphaStrong alphaStrongDef;
    AlphaStrong alphaStrongDefCMW;
    
    // Couplings for use in merging.
    AlphaStrong alphaS;
    AlphaEM     alphaEM;
    double mu2freeze, mu2min, alphaSmax;
    
    // Quark masses
    double ms, mc, mb, mt;
    
    // Checks
    double epTolErr, epTolWarn;
    double mTolErr, mTolWarn;

  private:

    // Pointers
    Settings*      settingsPtr; 
    ParticleData*  particleDataPtr;
    
    // Counter for output control
    int nUnkownPDG, nIncorrectCol, nNAN, nVertex, nChargeCons, nMotDau;
    vector<int> nUnmatchedMass, nEPcons;
    
    // Internal flags and settings
    bool isInitPtr, isInit;
    int verbose;

  };
 
  //**************************************************************************
  
  // Globally accessible functions and constants in the Vincia namespace
  
  // Color factors in Vincia normalization
  const double CA=3.0;
  const double CF=8.0/3.0;
  const double TR=1.0;
  const double NC=3.0;

  // Indices of FF antenna functions
  const int iQQemit  = 0;
  const int iQGemit  = 1;
  const int iGGemit  = 2;
  const int iQGsplit = 3;
  const int iGGsplit = 4;

  // Indices of II antenna functions
  const int iQQemitII  = 0;
  const int iGQemitII  = 1;
  const int iGGemitII  = 2;
  const int iQXsplitII = 3;
  const int iGXconvII  = 4;

  // Indices of IF antenna functions
  const int iQQemitIF  = 5;
  const int iQGemitIF  = 6;
  const int iGQemitIF  = 7;
  const int iGGemitIF  = 8;
  const int iQXsplitIF = 9;
  const int iGXconvIF  = 10;
  const int iXGsplitIF = 11;

  // Mathematical constants
  const double gammaE=0.577215664901532860606512090082402431042;
  
  // Max index for antenna coefficients
  const int LaurentIndexMax  = 10;
  // MR 10.7.2010: Max index for the invariants in the massive
  // antennae
  const int LaurentIndexMaxMassiveInv = 4;
  const int LaurentIndexMaxMassiveMass = 8;
  
  // A few useful auxiliary functions like extra invariants and dot products
  double m(const Vec4&);
  double m2(const Vec4&);
  double m2(const Vec4&, const Vec4&, const Vec4&);
  double m2(const Vec4&, const Vec4&, const Vec4&, const Vec4&);
  double m2(const Particle&, const Particle&, const Particle&);
  double dot4(const Vec4&, const Vec4&);
  double dot4(const Particle&, const Particle&);
  void Vec4toD4(const Vec4&, double*);
  void Vec4toMG(const Vec4&, double*);
  void Vec4toMG(const vector<Vec4>& v, int i0, int i1, int i2,
                double p[4][4]);
  void Vec4toMG(const vector<Vec4>& v, int i0, int i1, int i2, int i3, 
                double p[5][4]);
  void Vec4toMG(const vector<Vec4>& v, int i0, int i1, int i2, int i3, int i4, 
                double p[6][4]);
  void Vec4toMGQCD(const vector<Vec4>& v, int i0, int i1, int i2, int i3, 
                   double p[4][4]);
  void Vec4toMGQCD(const vector<Vec4>& v, int i0, int i1, int i2, int i3, int i4, 
                   double p[5][4]);
  void Vec4toMGQCD(const vector<Vec4>& v, int i0, int i1, int i2, int i3, int i4, int i5,
                   double p[6][4]);
  
  // Commen function for Pimp factor
  double PimpFac(double Q2hat, double Q2new, bool isVar = false);
  
  // Utilities to find daughters and ancestor of particles
  int findDaughters(Event&, int, vector<int>&);
  // Function to return whether parton i in event has iAncestor in its history
  bool isAncestor(Event& event, int i, int iAncestor);
  
  // Helper function to sort
  bool compVecSizeHelper(vector<int> a, vector<int> b);
  
  // Case insensitive comparison of two strings
  bool caseInsensCompPred(unsigned char a, unsigned char b);
  bool caseInsensComp(string const& a, string const& b);
  
  // Functors seem to be the only way to use zbrent for the rescaling 
  // variable needed in the massive Rambo version
  // TFunctor should be a template class, but the generality is 
  // not needed at this point
  
  // pure virtual function overloading operator()
  class TFunctor {
    public:
      virtual double operator()(double) =0;
  };
  
  class TXiFunctor : public TFunctor {
    public:
      // constructor takes a vector with the masses and one with the 
      // energies
      TXiFunctor(vector<double> mIn, vector<double> energiesIn);
      double operator() (double); 
    private:
      vector<double> m;
      vector<double> energies;
  };
  
  // derive a class which can wrap function pointers
  class TPtrFunctor : public TFunctor {
    public:
      TPtrFunctor(double (*fPtrIn)(double) ){ fPtr = fPtrIn; };
      double operator() (double arg) { return fPtr(arg); };
    private:
      // pointer to a function taking one double and returning one
      // double
      double (*fPtr)(double);
  };

  
  // Math support tools
  double zbrent(double (*)(double), double, double, double, double);
  // zbrentAlt is code duplication of zbrent, should replace zbrent, 
  // but testing is not complete
  // at present, doesn't work if the reference to TFunctor is declared
  // const
  double zbrentAlt(TFunctor&, double, double, double, double);
  double integrateGauss( double (*)(double), double, double, double);
  void makeSpline(vector<double>&, vector<double>&, vector<double>&);
  double getSpline(const double, vector<double>&, vector<double>&, 
                   vector<double>&);
  
  // Math support functions
  double Li2(const double,const double kmax=100.0,const double xerr=TINY);
  double lnGamma(const double);
  double Gamma(const double);
  int factorial(const int);
  int binomial(const int,int);
  
  // MR 21.12.2010: Gram determinant
  // the inviariants used in the argument are 2*pi*pj
  double gDet(
    double s12tilde, double s23tilde, double s13tilde, 
    double m1, double m2, double m3);
  // version with y13 omitted for normalised coordinates
  double gDet(
    double y12tilde, double y23tilde, 
    double mu1, double mu2,double mu3);
  double gDet(Vec4 p1, Vec4 p2, Vec4 p3);
    
  //**************************************************************************
  
  // A simple class to contain a spline
  
  class spline {
  public: 
    
    // constructor
    spline() {};
    
    // Members
    vector<double> x, y, d2;
  };
  
  //**************************************************************************

  // function to convert a vector of particle names to id's
  vector<int> nameVec2idVec(vector<string> nameVec);
  
  // From http://stackoverflow.com/questions/5343190
  inline string replaceString(string subject, const string& search,
                              const string& replace) {
    string::size_type pos = 0;
    while ((pos = subject.find(search, pos)) != string::npos) {
      subject.replace(pos, search.length(), replace);
      pos += replace.length();
    }
    return subject;
  }
  
  inline string sanitizeFileName(string fileName) {
    map<string, string> rep;
    rep["/"] = "_div_";
    rep[":"] = "_colon_";
    string retVal = fileName;
    for (map<string, string>::const_iterator it = rep.begin(); it != rep.end();
         ++it) {
      retVal = replaceString(retVal, it->first, it->second);
    }
    return retVal;
  }
  
} // End VINCIA namespace
#endif
